# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Based definition for authorization policies."""


class RuleNotForModelClasses(Exception):
    """Raised by authorization policies that don't make sense on model classes.

    For instance, if a permission check is done on a model class, and the
    policy associations are made with a policy that needs to check attributes,
    then the association must be corrected.
    """

    def __init__(self, policy, model):
        self.policy = policy
        self.model = model
        self.message = "Rule %r cannot be used on a model class (got %r)" % (
            policy, model)


class AuthorizationRule:
    """Base class to define the interface and provide some helpers"""

    registry = None
    """Set during assembly phase."""

    def is_declaration(self):
        return self.registry is None

    def check(self, target, principals, permission):
        """Check that one of the principals has permisson on given record.

        :param target: model instance (record) or class. Checking a permission
                       on a model class with a policy that is designed to work
                       on records is considered a configuration error,
                       expressed by :exc:`RuleNotForModelClasses`.
        :param principals: list, set or tuple of strings
        :rtype: bool

        Must be implemented by concrete subclasses.
        """
        raise NotImplementedError

    def filter(self, model, query, principals, permission):
        """Return a new query with added permission filtering.

        Must be implemented by concrete subclasses.

        :param query: the :class:`Query` instance to modify to express
                      the permission for these principals.
        :param model: the model on which the policy is applied
        :rtype: :class:`Query`)

        It's not necessary that the resulting query expresses fully
        the permission check: this can be complemented if needed
        by postfiltering, notably for conditions that can't be expressed
        conveniently in SQL.

        That being said, if the policy can be expressed totally by query,
        alteration, it's usually the best choice, as it keeps database traffic
        at the lowest.

        The policy also has the possibility to return False, for flat denial
        without even querying the server. That may prove useful in some cases.
        """
        raise NotImplementedError

    postfilter = None
    """Filter by permission records obtained by a filtered query.

    By default, this is ``None``, to indicate that the policy does not perform
    any post filtering, but concrete policies can implement
    a method with the following signature::

        def postfilter(self, record, principals, permission):

    Such implementations can (and usually, for performance, should) assume
    that the query that produced the records was a filtered one.

    The purpose of using the explicit ``None`` marker is to permit some calls
    that don't make sense on a postfiltered operation (such as ``count()``).
    """


class DenyAll(AuthorizationRule):

    def check(self, *args):
        return False

    def filter(self, *args):
        return False


deny_all = DenyAll



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from sqlalchemy.exc import IntegrityError
from anyblok.field import FieldException
from anyblok.column import (Integer,
                            String,
                            SmallInteger,
                            BigInteger,
                            Float,
                            Decimal,
                            uString,
                            Boolean,
                            DateTime,
                            Date,
                            Time,
                            Sequence)
from anyblok.relationship import Many2One
from sqlalchemy import ForeignKeyConstraint
from sqlalchemy.engine.reflection import Inspector


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


def _complete_many2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many="persons", nullable=False)


def _many2one_with_same_name_for_column_names(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           column_names="address")


def _minimum_many2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


def _many2one_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model='Model.Address')


def _minimum_many2one_without_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One()


def _auto_detect_type(ColumnType=None, **kwargs):

    @register(Model)
    class Address:

        id = ColumnType(primary_key=True, **kwargs)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


def _two_remote_primary_keys(**kwargs):

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = Many2One(model=Model.Test)


class TestMany2One(DBTestCase):

    def test_complete_many2one(self):
        registry = self.init_registry(_complete_many2one)
        address_exist = hasattr(registry.Person, 'address')
        self.assertTrue(address_exist)
        id_of_address_exist = hasattr(registry.Person, 'id_of_address')
        self.assertTrue(id_of_address_exist)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        self.assertEqual(address.persons, [person])

    def test_complete_many2one_expire_field(self):
        registry = self.init_registry(_complete_many2one)
        self.assertIn(
            ('address',),
            registry.expire_attributes['Model.Person']['id_of_address'])

    def test_2_many2one(self):
        with self.assertRaises(FieldException):
            self.init_registry(_many2one_with_same_name_for_column_names)

    def test_minimum_many2one(self):
        registry = self.init_registry(_minimum_many2one)
        address_exist = hasattr(registry.Person, 'address')
        self.assertTrue(address_exist)
        address_id_exist = hasattr(registry.Person, 'address_id')
        self.assertTrue(address_id_exist)
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        self.assertEqual(person.address, address)

    def test_minimum_many2one_on_sequence(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                seq = Sequence(primary_key=True, formater="V-{seq}")

            @register(Model)
            class Test2:

                seq = Sequence(primary_key=True)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        test2 = registry.Test2.insert(test=test)
        self.assertIs(test, test2.test)

    def test_minimum_many2one_expire_field(self):
        registry = self.init_registry(_minimum_many2one)
        self.assertEqual(
            registry.expire_attributes['Model.Person']['address_id'],
            {('address',)})

    def test_many2one_with_str_model(self):
        registry = self.init_registry(_many2one_with_str_model)
        address_exist = hasattr(registry.Person, 'address')
        self.assertTrue(address_exist)
        address_id_exist = hasattr(registry.Person, 'address_id')
        self.assertTrue(address_id_exist)
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        self.assertIs(person.address, address)

    def test_minimum_many2one_without_model(self):
        with self.assertRaises(FieldException):
            self.init_registry(_minimum_many2one_without_model)

    def check_autodetect_type(self, ColumnType):
        registry = self.init_registry(_auto_detect_type, ColumnType=ColumnType)
        self.assertEqual(
            str(registry.Address.id.property.columns[0].type),
            str(registry.Person.address_id.property.columns[0].type))

    def test_autodetect_type_integer(self):
        self.check_autodetect_type(Integer)

    def test_autodetect_type_small_integer(self):
        self.check_autodetect_type(SmallInteger)

    def test_autodetect_type_big_integer(self):
        self.check_autodetect_type(BigInteger)

    def test_autodetect_type_float(self):
        self.check_autodetect_type(Float)

    def test_autodetect_type_decimal(self):
        self.check_autodetect_type(Decimal)

    def test_autodetect_type_string(self):
        self.check_autodetect_type(String)

    def test_autodetect_type_ustring(self):
        self.check_autodetect_type(uString)

    def test_autodetect_type_boolean(self):
        self.check_autodetect_type(Boolean)

    def test_autodetect_type_datetime(self):
        self.check_autodetect_type(DateTime)

    def test_autodetect_type_date(self):
        self.check_autodetect_type(Date)

    def test_autodetect_type_time(self):
        self.check_autodetect_type(Time)

    def test_same_model(self):
        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertEqual(t2.parent, t1)
        self.assertEqual(t1.children[0], t2)

    def test_same_model_pk_by_inherit(self):
        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)  # noqa
            class Test:

                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertEqual(t2.parent, t1)
        self.assertEqual(t1.children[0], t2)

    def test_same_model_pk_by_mixin(self):
        def add_in_registry():

            @register(Mixin)
            class MTest:

                id = Integer(primary_key=True)

            @register(Model)  # noqa
            class Test(Mixin.MTest):

                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertEqual(t2.parent, t1)
        self.assertEqual(t1.children[0], t2)

    def test_add_unique_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, unique=True)

        registry = self.init_registry(add_in_registry)
        address = registry.Address.insert()
        registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        with self.assertRaises(IntegrityError):
            registry.Person.insert(name="Other", address=address)

    def test_add_index_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, index=True)

        registry = self.init_registry(add_in_registry)
        inspector = Inspector(registry.session.connection())
        indexes = inspector.get_indexes(registry.Person.__tablename__)
        self.assertEqual(len(indexes), 1)

    def test_add_primary_keys_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, primary_key=True)

        registry = self.init_registry(add_in_registry)
        inspector = Inspector(registry.session.connection())
        pks = inspector.get_primary_keys(registry.Person.__tablename__)
        self.assertIn('address_id', pks)

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                @classmethod
                def define_table_args(cls):
                    table_args = super(Test2, cls).define_table_args()
                    return table_args + (ForeignKeyConstraint(
                        [cls.test_id, cls.test_id2], ['test.id', 'test.id2']),)

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test,
                                remote_columns=('id', 'id2'),
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_complet_with_multi_foreign_key_without_constraint(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test,
                                remote_columns=('id', 'id2'),
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_existing_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_existing_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                other_test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                other_test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.other_test_id)
        self.assertEqual(test.id2, test2.other_test_id2)

    def test_with_multi_foreign_key_on_unexisting_column(self):
        registry = self.init_registry(_two_remote_primary_keys)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_many2one_with_multi_fk_expire_field(self):
        registry = self.init_registry(_two_remote_primary_keys)
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id'])
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id2'])

    def test_with_multi_foreign_key_on_unexisting_named_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test,
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_unexisting_named_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = Integer(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, column_names=(
                    'other_test_id', 'other_test_id2'))

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)

    def test_m2o_in_mixin(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Mixin)
            class MTest:

                test = Many2One(model=Model.Test)

            @register(Model)
            class Test1(Mixin.MTest):

                id = Integer(primary_key=True)

            @register(Model)
            class Test2(Mixin.MTest):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        test1 = registry.Test1.insert(test=test)
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test1.test_id, test2.test_id)

    def test_delete_m2o_without_fk_options_on_delete_cascade(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class TestM2O:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, nullable=True)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        testM2O = registry.TestM2O.insert(test=test)
        self.assertIs(testM2O.test, test)
        with self.assertRaises(IntegrityError):
            test.delete()

    def test_delete_m2o_with_fk_options_on_delete_cascade(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class TestM2O:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, nullable=True,
                                foreign_key_options={'ondelete': 'cascade'})

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        testM2O = registry.TestM2O.insert(test=test)
        self.assertIs(testM2O.test, test)
        test.delete()
        self.assertFalse(registry.TestM2O.query().count())

    def test_2_simple_many2one_on_the_same_model(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1 = Many2One(model=Model.Address)
                address_2 = Many2One(model=Model.Address)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1=address_1, address_2=address_2)
        self.assertIs(person.address_1, address_1)
        self.assertIs(person.address_2, address_2)
        self.assertIsNot(person.address_1, person.address_2)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok.model import ViewException
from anyblok import Declarations
from sqlalchemy.sql import select
from sqlalchemy.exc import OperationalError
from anyblok.column import Integer, String

register = Declarations.register
Model = Declarations.Model


def simple_view():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, is_sql_view=True)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.registry.T1
            T2 = cls.registry.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)


def simple_view_with_same_table_by_declaration_model():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, is_sql_view=True)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.registry.T1
            T2 = cls.registry.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)

    @register(Model, is_sql_view=True, tablename=Model.TestView)
    class TestView2:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


def simple_view_with_same_table_by_name():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, is_sql_view=True)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.registry.T1
            T2 = cls.registry.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)

    @register(Model, is_sql_view=True, tablename='testview')
    class TestView2:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


def simple_view_with_same_table_by_inherit():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, is_sql_view=True)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.registry.T1
            T2 = cls.registry.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T2.val.label('val2')])
            return query.where(T1.code == T2.code)

    @register(Model, is_sql_view=True)
    class TestView2(Model.TestView):
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


def simple_view_without_primary_key():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, is_sql_view=True)
    class TestView:
        code = String()
        val1 = Integer()
        val2 = Integer()

        @classmethod
        def sqlalchemy_view_declaration(cls):
            T1 = cls.registry.T1
            T2 = cls.registry.T2
            query = select([T1.code.label('code'),
                            T1.val.label('val1'),
                            T1.val.label('val2')])
            return query.where(T1.code == T2.code)


def simple_view_without_view_declaration():

    @register(Model)
    class T1:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model)
    class T2:
        id = Integer(primary_key=True)
        code = String()
        val = Integer()

    @register(Model, is_sql_view=True)
    class TestView:
        code = String(primary_key=True)
        val1 = Integer()
        val2 = Integer()


class TestView(DBTestCase):

    def test_simple_view(self):
        registry = self.init_registry(simple_view)
        registry.T1.insert(code='test1', val=1)
        registry.T2.insert(code='test1', val=2)
        registry.T1.insert(code='test2', val=3)
        registry.T2.insert(code='test2', val=4)
        TestView = registry.TestView
        v1 = TestView.query().filter(TestView.code == 'test1').first()
        v2 = TestView.query().filter(TestView.code == 'test2').first()
        self.assertEqual(v1.val1, 1)
        self.assertEqual(v1.val2, 2)
        self.assertEqual(v2.val1, 3)
        self.assertEqual(v2.val2, 4)

    def check_same_view(self, registry):
        registry.T1.insert(code='test1', val=1)
        registry.T2.insert(code='test1', val=2)
        registry.T1.insert(code='test2', val=3)
        registry.T2.insert(code='test2', val=4)
        TestView = registry.TestView
        TestView2 = registry.TestView2
        self.assertEqual(registry.TestView.__view__,
                         registry.TestView2.__view__)
        v1 = TestView.query().filter(TestView.code == 'test1').first()
        v2 = TestView.query().filter(TestView2.code == 'test1').first()
        self.assertEqual(v1.val1, v2.val1)
        self.assertEqual(v1.val2, v2.val2)

    def test_same_view_by_declaration_model(self):
        registry = self.init_registry(
            simple_view_with_same_table_by_declaration_model)
        self.check_same_view(registry)

    def test_same_view_by_name(self):
        registry = self.init_registry(simple_view_with_same_table_by_name)
        self.check_same_view(registry)

    def test_same_view_by_inherit(self):
        registry = self.init_registry(simple_view_with_same_table_by_inherit)
        self.check_same_view(registry)

    def test_view_update_method(self):
        registry = self.init_registry(simple_view)
        registry.T1.insert(code='test1', val=1)
        registry.T2.insert(code='test1', val=2)
        registry.T1.insert(code='test2', val=3)
        registry.T2.insert(code='test2', val=4)
        with self.assertRaises(OperationalError):
            registry.TestView.query().update({'val2': 3})

    def test_view_delete_method(self):
        registry = self.init_registry(simple_view)
        registry.T1.insert(code='test1', val=1)
        registry.T2.insert(code='test1', val=2)
        registry.T1.insert(code='test2', val=3)
        registry.T2.insert(code='test2', val=4)
        with self.assertRaises(OperationalError):
            registry.TestView.query().delete()

    def test_simple_view_without_primary_key(self):
        with self.assertRaises(ViewException):
            self.init_registry(simple_view_without_primary_key)

    def test_simple_view_without_view_declaration(self):
        with self.assertRaises(ViewException):
            self.init_registry(simple_view_without_view_declaration)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok, BlokManager
from anyblok.release import version


class AnyBlokIOXML(Blok):
    """ XML Importer / Exporter behaviour

    .. warning::

        Importer and Exporter are not implemented yet

    """
    version = version
    author = 'Suzanne Jean-Sébastien'
    logo = '../anyblok-logo_alpha_256.png'

    required = [
        'anyblok-io',
    ]

    def __init__(self, registry):
        super(AnyBlokIOXML, self).__init__(registry)
        if not BlokManager.has_importer('xml'):
            BlokManager.add_importer('xml', 'Model.IO.Importer.XML')

    @classmethod
    def import_declaration_module(cls):
        from . import importer  # noqa
        from . import exporter  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import importer
        reload(importer)
        from . import exporter
        reload(exporter)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .common import python_version

issix = False
try:
    from importlib import reload as reload_module
except ImportError:
    issix = True
    from six.moves import reload_module


def reload_wraper(module):
    if issix:
        module2reload = module

    elif python_version() == (3, 3):
        module2reload = module.__name__
    elif python_version() >= (3, 4):
        module2reload = module
    else:
        raise ImportManagerException(
            "Unknow action to do to reload module %r" %
            module.__name__)

    reload_module(module2reload)


def reload_module_if_blok_is_reloading(module):
    from anyblok.environment import EnvironmentManager

    if EnvironmentManager.get('reload', default=False):
        reload_wraper(module)


class ImportManagerException(AttributeError):
    """ Exception for Import Manager """


class ImportManager:
    """ Use to import the blok or reload the blok imports

        Add a blok and imports its modules::

            blok = ImportManager.add('my blok')
            blok.imports()

        Reload the modules of a blok::

            if ImportManager.has('my blok'):
                blok = ImportManager.get('my blok')
                blok.reload()
                # import the unimported module
    """
    modules = {}

    @classmethod
    def add(cls, blok):
        """ Store the blok so that we know which bloks to reload if needed

        :param blok: name of the blok to add
        :rtype: loader instance
        :exception: ImportManagerException
        """
        from anyblok.blok import BlokManager
        if cls.has(blok):
            return cls.get(blok)

        if not BlokManager.has(blok):
            raise ImportManagerException("Unexisting blok")

        loader = Loader(blok)
        cls.modules[blok] = loader
        return loader

    @classmethod
    def get(cls, blok):
        """ Return the module imported for this blok

        :param blok: name of the blok to add
        :rtype: loader instance
        :exception: ImportManagerException
        """
        if not cls.has(blok):
            raise ImportManagerException('Unexisting blok %r' % blok)
        return cls.modules[blok]

    @classmethod
    def has(cls, blok):
        """ Return True if the blok was imported

        :param blok: name of the blok to add
        :rtype: boolean
        """
        return blok in cls.modules


class Loader:

    def __init__(self, blok):
        self.blok = blok

    def imports(self):
        """ Imports modules and / or packages listed in the blok path"""
        from anyblok.blok import BlokManager
        from anyblok.registry import RegistryManager

        RegistryManager.init_blok(self.blok)
        b = BlokManager.get(self.blok)
        b.import_declaration_module()

    def reload(self):
        """ Reload all the imports for this module

        :exception: ImportManagerException
        """
        from anyblok.blok import BlokManager
        from anyblok.registry import RegistryManager
        from anyblok.environment import EnvironmentManager

        b = BlokManager.get(self.blok)
        if not hasattr(b, 'reload_declaration_module'):
            return

        try:
            EnvironmentManager.set('reload', True)
            RegistryManager.init_blok(self.blok)
            b.reload_declaration_module(reload_wraper)
        finally:
            EnvironmentManager.set('reload', False)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok.release import version


class AnyBlokIO(Blok):
    """
    In / Out tool's:

    * Formater: convert value 2 str or str 2 value in function of the field,
    * Importer: main model to define an import,
    * Exporter: main model to define an export,
    """
    version = version
    author = 'Suzanne Jean-Sébastien'
    logo = '../anyblok-logo_alpha_256.png'

    required = [
        'anyblok-core',
    ]

    @classmethod
    def declare_io(cls):
        from anyblok import Declarations

        @Declarations.register(Declarations.Model)
        class IO:
            pass

    @classmethod
    def import_declaration_module(cls):
        from . import core  # noqa
        cls.declare_io()
        from . import mapping  # noqa
        from . import mixin  # noqa
        from . import importer  # noqa
        from . import exporter  # noqa
        from . import formater  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import core
        reload(core)
        cls.declare_io()
        from . import mapping
        reload(mapping)
        from . import mixin
        reload(mixin)
        from . import importer
        reload(importer)
        from . import exporter
        reload(exporter)
        from . import formater
        reload(formater)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.model import ViewException
from .sqlbase import SqlMixin


@Declarations.register(Declarations.Core)
class SqlViewBase(SqlMixin):
    """ this class is inherited by all the SQL view
    """

    def update(self, *args, **kwargs):
        raise ViewException("%r.update method are not availlable on view "
                            "model" % self.__registry_name__)

    def delete(self, *args, **kwargs):
        raise ViewException("%r.delete method are not availlable on view "
                            "model" % self.__registry_name__)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.registry import RegistryManager
from anyblok.blok import BlokManager
from anyblok.model import Model
from anyblok.environment import EnvironmentManager


class TestRegistryManager(TestCase):

    @classmethod
    def tearDownClass(cls):
        super(TestRegistryManager, cls).tearDownClass()
        RegistryManager.undeclare_entry('Other')

    def test_declared_entries(self):
        hasModel = 'Model' in RegistryManager.declared_entries
        hasMixin = 'Mixin' in RegistryManager.declared_entries
        self.assertEqual(hasModel, True)
        self.assertEqual(hasMixin, True)

    def test_init_blok(self):
        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        self.assertEqual(is_exist, True)
        for core in ('Base', 'SqlBase', 'SqlViewBase', 'Session', 'Query',
                     'InstrumentedList'):
            self.assertIn(
                core, RegistryManager.loaded_bloks['newblok']['Core'].keys())
            self.assertEqual(
                RegistryManager.loaded_bloks['newblok']['Core'][core], [])

        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Model'],
                         {'registry_names': []})
        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Mixin'],
                         {'registry_names': []})

    def test_init_blok_with_other_entry(self):
        RegistryManager.declare_entry('Other')
        hasOther = 'Other' in RegistryManager.declared_entries
        self.assertEqual(hasOther, True)
        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        self.assertEqual(is_exist, True)
        for core in ('Base', 'SqlBase', 'SqlViewBase', 'Session', 'Query',
                     'InstrumentedList'):
            self.assertIn(
                core, RegistryManager.loaded_bloks['newblok']['Core'].keys())
            self.assertEqual(
                RegistryManager.loaded_bloks['newblok']['Core'][core], [])

        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Model'],
                         {'registry_names': []})
        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Mixin'],
                         {'registry_names': []})
        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Other'],
                         {'registry_names': []})

    def test_anyblok_core_loaded(self):
        BlokManager.load()
        is_exist = 'anyblok-core' in RegistryManager.loaded_bloks
        self.assertEqual(is_exist, True)
        anyblokcore = RegistryManager.loaded_bloks['anyblok-core']
        self.assertEqual(len(anyblokcore['Core']['Base']), 1)
        self.assertEqual(len(anyblokcore['Core']['SqlBase']), 1)
        self.assertEqual(len(anyblokcore['Core']['SqlViewBase']), 1)
        self.assertEqual(len(anyblokcore['Core']['Session']), 1)
        self.assertEqual(len(anyblokcore['Core']['Query']), 1)
        self.assertEqual(len(anyblokcore['Core']['InstrumentedList']), 1)
        is_exist = 'Model.System' in anyblokcore['Model']
        self.assertEqual(is_exist, True)
        BlokManager.unload()

    def test_add_entry(self):
        RegistryManager.declare_entry('Other')
        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        self.assertEqual(is_exist, True)
        for entry in ('Base', 'SqlBase', 'SqlViewBase', 'Session', 'Query',
                      'InstrumentedList'):
            self.assertEqual(
                RegistryManager.loaded_bloks['newblok']['Core'][entry], [])

        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Model'],
                         {'registry_names': []})
        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Mixin'],
                         {'registry_names': []})
        self.assertEqual(RegistryManager.loaded_bloks['newblok']['Other'],
                         {'registry_names': []})

    def test_add_callback(self):

        def callback():
            pass

        RegistryManager.declare_entry('Other', assemble_callback=callback,
                                      initialize_callback=callback)
        hasModel = 'Model' in RegistryManager.declared_entries
        hasMixin = 'Mixin' in RegistryManager.declared_entries
        hasOther = 'Other' in RegistryManager.declared_entries
        self.assertEqual(hasModel, True)
        self.assertEqual(hasMixin, True)
        self.assertEqual(hasOther, True)

        cb = Model.assemble_callback
        hasModelCb = cb == RegistryManager.callback_assemble_entries['Model']
        cb = callback
        hasOtherCb = cb == RegistryManager.callback_assemble_entries['Other']
        self.assertEqual(hasModelCb, True)
        self.assertEqual(hasOtherCb, True)

        cb = Model.initialize_callback
        hasModelCb = cb == RegistryManager.callback_initialize_entries['Model']
        cb = callback
        hasOtherCb = cb == RegistryManager.callback_initialize_entries['Other']
        self.assertEqual(hasModelCb, True)
        self.assertEqual(hasOtherCb, True)

        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        self.assertEqual(is_exist, True)
        hasCore = 'Core' in RegistryManager.loaded_bloks['newblok']
        hasModel = 'Model' in RegistryManager.loaded_bloks['newblok']
        hasMixin = 'Mixin' in RegistryManager.loaded_bloks['newblok']
        hasOther = 'Other' in RegistryManager.loaded_bloks['newblok']
        self.assertEqual(hasCore, True)
        self.assertEqual(hasModel, True)
        self.assertEqual(hasMixin, True)
        self.assertEqual(hasOther, True)

    def test_reload_blok(self):
        BlokManager.load()
        try:
            RegistryManager.reload()
        finally:
            BlokManager.unload()

    def test_global_property(self):
        RegistryManager.declare_entry('Other')
        blok = 'newblok'
        RegistryManager.init_blok(blok)
        try:
            oldblok = EnvironmentManager.get('current_blok')
            EnvironmentManager.set('current_blok', blok)
            self.assertEqual(RegistryManager.has_blok_property('myproperty'),
                             False)
            RegistryManager.add_or_replace_blok_property('myproperty', 2)
            self.assertEqual(
                RegistryManager.has_blok_property('myproperty'), True)
            self.assertEqual(
                RegistryManager.get_blok_property('myproperty'), 2)
            RegistryManager.add_or_replace_blok_property('myproperty', 3)
            self.assertEqual(
                RegistryManager.get_blok_property('myproperty'), 3)
            RegistryManager.remove_blok_property('myproperty')
            self.assertEqual(RegistryManager.has_blok_property('myproperty'),
                             False)
        finally:
            EnvironmentManager.set('current_blok', oldblok)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from json import dumps


class TestExporterFormater(BlokTestCase):

    def get_value(self, value, ctype, external_id=False, model=None):
        return self.registry.IO.Exporter().value2str(value, ctype,
                                                     external_id=external_id,
                                                     model=model)

    def test_datetime(self):
        from datetime import datetime
        dt = datetime(2015, 5, 1, 17, 52, 1)
        value = self.get_value(dt, "DateTime")
        self.assertEqual(value, "2015-05-01 17:52:01")

    def test_decimal(self):
        from decimal import Decimal
        value = self.get_value(Decimal("1.5"), "Decimal")
        self.assertEqual(value, "1.5")

    def test_float(self):
        value = self.get_value(1.5, "Float")
        self.assertEqual(value, "1.5")

    def test_json(self):
        value = self.get_value({'id': 1}, "Json")
        self.assertEqual(value, dumps({'id': 1}))

    def test_time(self):
        from datetime import time
        t = time(17, 52, 1)
        value = self.get_value(t, "Time")
        self.assertEqual(value, "17:52:01")

    def test_big_interger(self):
        value = self.get_value(17, "BigInteger")
        self.assertEqual(value, "17")

    def test_boolean(self):
        value = self.get_value(True, "Boolean")
        self.assertEqual(value, '1')
        value = self.get_value(False, "Boolean")
        self.assertEqual(value, '0')

    def test_date(self):
        from datetime import date
        d = date(2015, 5, 1)
        value = self.get_value(d, "Date")
        self.assertEqual(value, "2015-05-01")

    def test_interger(self):
        value = self.get_value(10, "Integer")
        self.assertEqual(value, "10")

    def test_interval(self):
        from datetime import timedelta
        td = timedelta(seconds=10)
        value = self.get_value(td, "Interval")
        self.assertEqual(value, "10")

    def test_selection(self):
        value = self.get_value("selection", "Selection")
        self.assertEqual(value, "selection")

    def test_small_interger(self):
        value = self.get_value(10, "SmallInteger")
        self.assertEqual(value, "10")

    def test_string(self):
        value = self.get_value("selection", "String")
        self.assertEqual(value, "selection")

    def test_text(self):
        value = self.get_value("selection", "Text")
        self.assertEqual(value, "selection")

    def test_ustring(self):
        value = self.get_value("selection", "uString")
        self.assertEqual(value, "selection")

    def test_utext(self):
        value = self.get_value("selection", "uText")
        self.assertEqual(value, "selection")

    def test_many2many(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value([model], "Many2Many",
                               model="Model.System.Model")
        self.assertEqual(value, dumps([pks]))

    def test_many2many_multi_value(self):
        pks1 = dict(name='Model.System.Model')
        pks2 = dict(name='Model.System.Column')
        model1 = self.registry.System.Model.from_primary_keys(**pks1)
        model2 = self.registry.System.Model.from_primary_keys(**pks2)
        value = self.get_value([model1, model2], "Many2Many",
                               model="Model.System.Model")
        self.assertEqual(value, dumps([pks1, pks2]))

    def test_many2many_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(([model]), "Many2Many", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, dumps([key]))

    def test_many2many_multi_value_external_ids(self):
        key1 = 'formater_mapping1'
        key2 = 'formater_mapping2'
        model1 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        model2 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Column')
        self.registry.IO.Mapping.set(key1, model1)
        self.registry.IO.Mapping.set(key2, model2)
        value = self.get_value([model1, model2], "Many2Many",
                               external_id=True, model="Model.System.Model")
        self.assertEqual(value, dumps([key1, key2]))

    def test_one2many(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value([model], "One2Many", model="Model.System.Model")
        self.assertEqual(value, dumps([pks]))

    def test_one2many_multi_value(self):
        pks1 = dict(name='Model.System.Model')
        pks2 = dict(name='Model.System.Column')
        model1 = self.registry.System.Model.from_primary_keys(**pks1)
        model2 = self.registry.System.Model.from_primary_keys(**pks2)
        value = self.get_value([model1, model2], "One2Many",
                               model="Model.System.Model")
        self.assertEqual(value, dumps([pks1, pks2]))

    def test_one2many_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value([model], "One2Many", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, dumps([key]))

    def test_one2many_multi_value_external_ids(self):
        key1 = 'formater_mapping1'
        key2 = 'formater_mapping2'
        model1 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        model2 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Column')
        self.registry.IO.Mapping.set(key1, model1)
        self.registry.IO.Mapping.set(key2, model2)
        value = self.get_value([model1, model2], "One2Many",
                               external_id=True, model="Model.System.Model")
        self.assertEqual(value, dumps([key1, key2]))

    def test_many2one(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value(model, "Many2One", model="Model.System.Model")
        self.assertEqual(value, dumps(pks))

    def test_many2one_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(model, "Many2One", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, key)

    def test_one2one(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value(model, "One2One", model="Model.System.Model")
        self.assertEqual(value, dumps(pks))

    def test_one2one_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(model, "One2One", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, key)

    def test_external_id(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(model.name, "String", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, key)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.declarations import Declarations, hybrid_method
from anyblok.column import String, Json
from .exceptions import IOMappingCheckException, IOMappingSetException
from logging import getLogger
logger = getLogger(__name__)


register = Declarations.register
Model = Declarations.Model


@register(Model.IO)
class Mapping:

    key = String(primary_key=True)
    model = String(primary_key=True,
                   foreign_key=Model.System.Model.use('name'))
    primary_key = Json(nullable=False)
    blokname = String(label="Blok name",
                      foreign_key=Model.System.Blok.use('name'))

    @hybrid_method
    def filter_by_model_and_key(self, model, key):
        """ SQLAlechemy hybrid method to filter by model and key

        :param model: model of the mapping
        :param key: external key of the mapping
        """
        return (self.model == model) & (self.key == key)

    @hybrid_method
    def filter_by_model_and_keys(self, model, *keys):
        """ SQLAlechemy hybrid method to filter by model and key

        :param model: model of the mapping
        :param key: external key of the mapping
        """
        return (self.model == model) & self.key.in_(keys)

    def remove_element(self, byquery=False):
        val = self.registry.get(self.model).from_primary_keys(
            **self.primary_key)
        logger.info("Remove entity for %r.%r: %r" % (
            self.model, self.key, val))
        val.delete(byquery=byquery, remove_mapping=False)

    @classmethod
    def multi_delete(cls, model, *keys, **kwargs):
        """ Delete all the keys for this model

        :param model: model of the mapping
        :param \*keys: list of the key
        :rtype: Boolean True if the mappings are removed
        """
        mapping_only = kwargs.get('mapping_only', True)
        byquery = kwargs.get('byquery', False)
        query = cls.query()
        query = query.filter(cls.filter_by_model_and_keys(model, *keys))
        count = query.count()
        if count:
            if not mapping_only:
                query.all().remove_element(byquery=byquery)

            query.delete(synchronize_session='fetch', remove_mapping=False)
            cls.registry.expire_all()
            return count

        return 0

    @classmethod
    def delete(cls, model, key, mapping_only=True, byquery=False):
        """ Delete the key for this model

        :param model: model of the mapping
        :param key: string of the key
        :rtype: Boolean True if the mapping is removed
        """
        query = cls.query()
        query = query.filter(cls.filter_by_model_and_key(model, key))
        count = query.count()
        if count:
            if not mapping_only:
                query.one().remove_element(byquery=byquery)

            query.delete(remove_mapping=False)
            return count

        return 0

    @classmethod
    def get_mapping_primary_keys(cls, model, key):
        """ return primary key for a model and an external key

        :param model: model of the mapping
        :param key: string of the key
        :rtype: dict primary key: value or None
        """
        query = cls.query()
        query = query.filter(cls.filter_by_model_and_key(model, key))
        if query.count():
            pks = query.first().primary_key
            cls.check_primary_keys(model, *pks.keys())
            return pks

        return None

    @classmethod
    def check_primary_keys(cls, model, *pks):
        """ check if the all the primary keys match with primary keys of the
        model

        :param model: model to check
        :param pks: list of the primary keys to check
        :exception: IOMappingCheckException
        """
        for pk in cls.get_model(model).get_primary_keys():
            if pk not in pks:
                raise IOMappingCheckException(
                    "No primary key %r found in %r for model %r" % (
                        pk, pks, model))

    @classmethod
    def set_primary_keys(cls, model, key, pks, raiseifexist=True,
                         blokname=None):
        """ Add or update a mmping with a model and a external key

        :param model: model to check
        :param key: string of the key
        :param pks: dict of the primary key to save
        :param raiseifexist: boolean (True by default), if True and the entry
            exist then an exception is raised
        :param blokname: name of the blok where come from the mapping
        :exception: IOMappingSetException
        """
        if cls.get_mapping_primary_keys(model, key):
            if raiseifexist:
                raise IOMappingSetException(
                    "One value found for model %r and key %r" % (model, key))
            cls.delete(model, key)

        if not pks:
            raise IOMappingSetException(
                "No value to save %r for model %r and key %r" % (
                    pks, model, key))

        cls.check_primary_keys(model, *pks.keys())
        vals = dict(model=model, key=key, primary_key=pks)
        if blokname is not None:
            vals['blokname'] = blokname

        return cls.insert(**vals)

    @classmethod
    def set(cls, key, instance, raiseifexist=True, blokname=None):
        """ Add or update a mmping with a model and a external key

        :param model: model to check
        :param key: string of the key
        :param instance: instance of the model to save
        :param raiseifexist: boolean (True by default), if True and the entry
            exist then an exception is raised
        :param blokname: name of the blok where come from the mapping
        :exception: IOMappingSetException
        """
        pks = instance.to_primary_keys()
        return cls.set_primary_keys(instance.__registry_name__, key, pks,
                                    blokname=blokname,
                                    raiseifexist=raiseifexist)

    @classmethod
    def get(cls, model, key):
        """ return instance of the model with this external key

        :param model: model of the mapping
        :param key: string of the key
        :rtype: instance of the model
        """
        pks = cls.get_mapping_primary_keys(model, key)
        if pks is None:
            return None

        return cls.get_model(model).from_primary_keys(**pks)

    @classmethod
    def get_from_model_and_primary_keys(cls, model, pks):
        query = cls.query().filter(cls.model == model)
        for mapping in query.all():
            if mapping.primary_key == pks:
                return mapping

        return None

    @classmethod
    def get_from_entry(cls, entry):
        return cls.get_from_model_and_primary_keys(
            entry.__registry_name__, entry.to_primary_keys())

    @classmethod
    def __get_models(cls, models):
        """Return models name

        if models is not: return all the existing model
        if models is a list of instance model, convert them

        :params models: list of model
        """
        if models is None:
            models = cls.registry.System.Model.query().all().name
        elif not isinstance(models, (list, tuple)):
            models = [models]

        return [m.__registry_name__ if hasattr(m, '__registry_name__') else m
                for m in models]

    @classmethod
    def clean(cls, bloknames=None, models=None):
        """Clean all mapping with removed object linked::

            Mapping.clean(bloknames=['My blok'])

        .. warning::

            For filter only the no blokname::

                Mapping.clean(bloknames=[None])

        :params bloknames: filter by blok, keep the order to remove the mapping
        :params models: filter by model, keep the order to remove the mapping
        """
        if bloknames is None:
            bloknames = cls.registry.System.Blok.query().all().name + [None]
        elif not isinstance(bloknames, (list, tuple)):
            bloknames = [bloknames]
        models = cls.__get_models(models)

        removed = 0
        for blokname in bloknames:
            for model in models:
                query = cls.query().filter_by(blokname=blokname, model=model)
                for key in query.all().key:
                    if cls.get(model, key) is None:
                        cls.delete(model, key)
                        removed += 1

        return removed

    @classmethod
    def delete_for_blokname(cls, blokname, models=None, byquery=False):
        """Clean all mapping with removed object linked::

            Mapping.clean('My blok')

        .. warning::

            For filter only the no blokname::

                Mapping.clean(None)

        :params blokname: filter by blok
        :params models: filter by model, keep the order to remove the mapping
        """
        models = cls.__get_models(models)

        removed = 0
        for model in models:
            query = cls.query().filter_by(blokname=blokname, model=model)
            for key in query.all().key:
                if cls.get(model, key):
                    cls.delete(model, key, mapping_only=False, byquery=byquery)
                    cls.registry.flush()
                    removed += 1

        return removed



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from logging import getLogger
logger = getLogger(__name__)


@Declarations.register(Declarations.Model.Documentation.Model)
class Field:

    mappers = {
        ('Many2One', True): ("m2o", "o2m"),
        ('Many2One', False): ("m2o", None),
        ('Many2Many', True): ("m2m", "m2m"),
        ('Many2Many', False): ("m2m", None),
        ('One2Many', True): ("o2m", "m2o"),
        ('One2Many', False): ("o2m", None),
        ('One2One', True): ("o2o", "o2o"),
        ('One2One', False): ("o2o", "o2o"),
    }

    def exist(self, model):
        if not model.exist():
            return False

        M = self.registry.get(model.model.name)
        if self.field.name in M.loaded_columns:
            return True

        return False

    def __init__(self, field):
        self.field = field

    @classmethod
    def filterField(cls, query):
        return query

    @classmethod
    def getelements(cls, model):
        query = cls.filterField(cls.registry.System.Field.query())
        return query.filter(
            cls.registry.System.Field.model == model.model.name).all()

    @classmethod
    def header2RST(cls, doc):
        doc.write("Fields\n~~~~~~\n\n")

    @classmethod
    def footer2RST(cls, doc):
        pass

    def toRST(self, doc):
        doc.write('* ' + self.field.name + '\n\n')
        self.toRST_docstring(doc)
        self.toRST_properties(doc)

    def toRST_docstring(self, doc):
        if hasattr(self.field, '__doc__') and self.field.__doc__:
            doc.write(self.field.__doc__ + '\n\n')

    def toRST_properties_get(self):
        return {x: y for x, y in self.field.to_dict().items() if x != 'name'}

    def toRST_properties(self, doc):
        properties = self.toRST_properties_get()
        msg = ', '.join(' **%s** (%s)' % (x, y) for x, y in properties.items())
        doc.write(msg + '\n\n')

    def toUML(self, dot):
        if self.field.entity_type == 'Model.System.Field':
            self.toUML_field(dot)
        elif self.field.entity_type == 'Model.System.Column':
            self.toUML_column(dot)
        elif self.field.entity_type == 'Model.System.RelationShip':
            self.toUML_relationship(dot)
        else:
            logger.warn("Unknown entity type %r" % self.field.entity_type)

    def toUML_field(self, dot):
        model = dot.get_class(self.field.model)
        model.add_column(self.field.name)

    def toUML_column(self, dot):
        model = dot.get_class(self.field.model)
        name = self.field.name
        if self.field.primary_key:
            name = '+PK+ ' + name

        if self.field.remote_model:
            remote_model = dot.get_class(self.field.remote_model)
            multiplicity = "1"
            if self.field.nullable:
                multiplicity = "0..1"

            model.agregate(remote_model, label_from=name,
                           multiplicity_from=multiplicity)
        else:
            name += ' (%s)' % self.field.ftype
            model.add_column(name)

    def toUML_relationship(self, dot):
        if self.field.remote:
            return

        model = dot.get_class(self.field.model)
        multiplicity, multiplicity_to = self.mappers[(
            self.field.ftype, True if self.field.remote_name else False)]
        model.associate(self.field.remote_model, label_from=self.field.name,
                        label_to=self.field.remote_name,
                        multiplicity_from=multiplicity,
                        multiplicity_to=multiplicity_to)

    def toSQL(self, dot):
        if self.field.entity_type == 'Model.System.Field':
            self.toSQL_field(dot)
        elif self.field.entity_type == 'Model.System.Column':
            self.toSQL_column(dot)
        elif self.field.entity_type == 'Model.System.RelationShip':
            self.toSQL_relationship(dot)
        else:
            logger.warn("Unknown entity type %r" % self.field.entity_type)

    def toSQL_field(self, dot):
        # DO NOTHING
        pass

    def toSQL_relationship(self, dot):
        # TODO
        pass

    def toSQL_column(self, dot):
        table = dot.get_table(self.registry.get(
            self.field.model).__tablename__)
        if self.field.foreign_key:
            remote_table = dot.get_table(self.field.foreign_key.split('.')[0])
            if remote_table is None:
                remote_table = dot.add_label(
                    self.field.foreign_key.split('.')[0])

            table.add_foreign_key(remote_table, label=self.field.name,
                                  nullable=self.field.nullable)
        else:
            table.add_column(self.field.name, self.field.ftype,
                             primary_key=self.field.primary_key)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations


@Declarations.register(Declarations.Core)
class InstrumentedList:
    """ class of the return of the query.all() or the relationship list
    """

    def __getattr__(self, name):
        if name in ('__emulates__', ):
            return None

        def wrapper(*args, **kwargs):
            return [getattr(x, name)(*args, **kwargs) for x in self]

        if not self:
            return []
        elif hasattr(getattr(self[0], name), '__call__'):
            return wrapper
        else:
            return [getattr(x, name) for x in self]



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String, Integer


register = Declarations.register
Mixin = Declarations.Mixin


@register(Mixin)
class IOCSVFieldMixin:

    id = Integer(primary_key=True)
    name = String(nullable=False)


@register(Mixin)
class IOCSVMixin:

    csv_delimiter = String(nullable=False, default=",")
    csv_quotechar = String(nullable=False, default='"')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok.release import version
from logging import getLogger
logger = getLogger(__name__)


class AnyBlokCore(Blok):
    """
    This blok is required by all anyblok application. This blok define the main
    fonctionnality to install, update and uninstall blok. And also list the
    known models, fields, columns and relationships
    """

    version = version
    autoinstall = True
    priority = 0
    author = 'Suzanne Jean-Sébastien'
    logo = '../anyblok-logo_alpha_256.png'

    def pre_migration(self, latest_version):
        if latest_version is not None and latest_version < '0.4.1':
            self.pre_migration_0_4_1_fields_become_polymorphic(latest_version)

    def pre_migration_0_4_1_fields_become_polymorphic(self, latest_version):
        logger.info("Pre Migration %s => %s: Field, Column, Relation Ship "
                    "become prolymophic models" % (latest_version,
                                                   self.version))
        system_field = self.registry.migration.table('system_field')
        system_field.column().add(self.registry.System.Field.entity_type)
        self.registry.execute(
            "UPDATE system_field SET entity_type='Model.System.Field'")
        query = """
            INSERT INTO system_field (
                name,
                model,
                code,
                label,
                ftype,
                entity_type)
            SELECT
                name,
                model,
                code,
                label,
                ftype,
                '%(entity_type)s' AS entity_type
            FROM %(table)s
        """
        self.registry.execute(
            query % {'entity_type': 'Model.System.Column',
                     'table': 'system_column'})
        self.registry.execute(
            query % {'entity_type': 'Model.System.RelationShip',
                     'table': 'system_relationship'})
        system_column = self.registry.migration.table('system_column')
        system_column.column('code').drop()
        system_column.column('ftype').drop()
        system_column.column('label').drop()
        system_relationship = self.registry.migration.table(
            'system_relationship')
        system_relationship.column('code').drop()
        system_relationship.column('ftype').drop()
        system_relationship.column('label').drop()

    @classmethod
    def import_declaration_module(cls):
        from . import core  # noqa
        from . import system  # noqa
        from . import authorization  # noqa
        from . import documentation  # noqa
        from . import mixins  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import core
        reload(core)
        from . import system
        reload(system)
        from . import authorization
        reload(authorization)
        from . import documentation
        reload(documentation)
        from . import mixins
        reload(mixins)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations


@Declarations.register(Declarations.Core)
class Query:

    def delete(self, *args, **kwargs):
        """Inherit the Query.delete methods.::

            Model.query().delete(remove_mapping=True)

        :param remove_mapping: boolean, if check (default) the mapping is
            removed
        """
        remove_mapping = kwargs.pop('remove_mapping', True)
        if remove_mapping:
            Mapping = self.registry.IO.Mapping
            mappings = []
            for entry in self.all():
                mapping = Mapping.get_from_model_and_primary_keys(
                    entry.__registry_name__, entry.to_primary_keys())
                if mapping and mapping not in mappings:
                    mappings.append(mapping)

            res = super(Query, self).delete(*args, **kwargs)
            for mapping in mappings:
                Mapping.delete(mapping.model, mapping.key)

            return res

        return super(Query, self).delete(*args, **kwargs)


@Declarations.register(Declarations.Core)
class SqlBase:

    def delete(self, *args, **kwargs):
        """Inherit the Model.delete methods.::

            instance.delete(remove_mapping=True)

        :param remove_mapping: boolean, if check (default) the mapping is
            removed
        """
        remove_mapping = kwargs.pop('remove_mapping', True)
        if remove_mapping:
            Mapping = self.registry.IO.Mapping
            mapping = Mapping.get_from_model_and_primary_keys(
                self.__registry_name__, self.to_primary_keys())

            res = super(SqlBase, self).delete(*args, **kwargs)
            if mapping:
                Mapping.delete(mapping.model, mapping.key)

            return res

        return super(SqlBase, self).delete(*args, **kwargs)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from .exceptions import ExporterException


@Declarations.register(Declarations.Model.IO)
class Exporter(Declarations.Mixin.IOMixin):

    def run(self, entries):
        for entry in entries:
            if entry.__registry_name__ != self.model:
                raise ExporterException(
                    "The entries must be instance of %r" % self.model)

        return self.get_model(self.mode)(self).run(entries)

    @classmethod
    def get_external_id(cls, model):
        Sequence = cls.registry.System.Sequence
        seq_code = 'export.%s' % model
        query = Sequence.query().filter(Sequence.code == seq_code)
        if query.count():
            sequence = query.first()
        else:
            sequence = Sequence.insert(formater="{code}_{seq}", code=seq_code)

        return sequence.nextval()

    @classmethod
    def get_key_mapping(cls, entry):
        Mapping = cls.registry.IO.Mapping
        mapping = Mapping.get_from_entry(entry)
        if mapping:
            return mapping.key

        key = cls.get_external_id(entry.__registry_name__)
        Mapping.set(key, entry)
        return key

    def value2str(self, value, ctype, external_id=False, model=None):
        formater = self.get_formater(ctype)
        if external_id:
            return formater.externalIdValue2str(value, model)

        return formater.value2str(value, model)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .plugins import ModelPluginBase
from anyblok.column import DateTime
from anyblok.mapper import ModelMapper
from datetime import datetime


class AutoUpdatePlugin(ModelPluginBase):

    def initialisation_tranformation_properties(self, properties,
                                                transformation_properties):
        """ Initialise the transform properties: auto_update_field_datetime

        :param properties: the properties declared in the model
        :param new_type_properties: param to add in a new base if need
        """
        if 'auto_update_field_datetime' not in transformation_properties:
            transformation_properties['auto_update_field_datetime'] = []

    def declare_field(self, name, field, namespace, properties,
                      transformation_properties):
        """Detect if the field is a DateTime with auto_update = True

        :param name: field name
        :param field: field instance
        :param namespace: the namespace of the model
        :param properties: the properties of the model
        :param transformation_properties: the transformation properties
        """
        f = self.registry.loaded_namespaces_first_step[namespace].get(name)
        if f and isinstance(f, DateTime) and f.auto_update:
            transformation_properties['auto_update_field_datetime'].append(
                name)

    def after_model_construction(self, base, namespace,
                                 transformation_properties):
        """Add the sqlalchemy event

        :param base: the Model class
        :param namespace: the namespace of the model
        :param transformation_properties: the properties of the model
        """
        if transformation_properties['auto_update_field_datetime']:
            e = ModelMapper(namespace, 'after_update')
            fields = transformation_properties['auto_update_field_datetime']

            def auto_update_listen(mapper, connection, target):
                now = datetime.now()
                for field in fields:
                    setattr(target, field, now)

            self.registry._sqlalchemy_known_events.append(
                (e, namespace, auto_update_listen))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import sys
from functools import lru_cache
from sqlalchemy.schema import ForeignKeyConstraint


"""Define the prefixe for the mapper attribute for the column"""
anyblok_column_prefix = '__anyblok_field_'


def all_column_name(constraint, table):
    """Define the convention for merge the column"""
    if isinstance(constraint, ForeignKeyConstraint):
        return '_'.join(constraint.column_keys)
    else:
        return '_'.join(constraint.columns.keys())


def model_name(constraint, table):
    """return a shortest table name"""
    name = table.name.split('_')
    if len(name) == 1:
        return name[0]

    return ''.join(x[0] for x in name[:-1]) + '_' + name[-1]


"""table convention for constraint"""
naming_convention = {
    "all_column_name": all_column_name,
    "model_name": model_name,
    "ix": "anyblok_ix_%(model_name)s__%(all_column_name)s",
    "uq": "anyblok_uq_%(model_name)s__%(all_column_name)s",
    "ck": "anyblok_ck_%(model_name)s__%(constraint_name)s",
    "fk": "anyblok_fk_%(model_name)s__%(all_column_name)s",
    "pk": "anyblok_pk_%(table_name)s",
}


def add_autodocs(meth, autodoc):
    if not hasattr(meth, 'autodocs'):
        meth.autodocs = []

    meth.autodocs.append(autodoc)


def function_name(function):
    if sys.version_info < (3, 3):
        return function.__name__
    else:
        return function.__qualname__


def python_version():
    vi = sys.version_info
    return (vi.major, vi.minor)


class TypeList(list):

    def __init__(self, Model, registry, namespace,
                 transformation_properties=None):
        super(TypeList, self).__init__()
        self.Model = Model
        self.registry = registry
        self.namespace = namespace
        self.transformation_properties = transformation_properties

    def transform_base(self, base, namespace=None):
        if base in self.registry.removed:
            return None

        if namespace is None:
            namespace = self.namespace

        newbase = self.Model.transform_base(
            self.registry, namespace, base, self.transformation_properties)
        return newbase

    def append(self, base, **kwargs):
        bases = self.transform_base(base, **kwargs) or []
        for newbase in bases:
            super(TypeList, self).append(newbase)

    def extend(self, bases, **kwargs):
        newbases = []
        for base in bases:
            _bases = self.transform_base(base, **kwargs)
            if _bases:
                newbases.extend(_bases)

        if newbases:
            super(TypeList, self).extend(newbases)


def apply_cache(attr, method, registry, namespace, base, properties):
    """ Find the cached methods in the base to apply the real cache
    decorator

    :param attr: name of the attibute
    :param method: method pointer
    :param registry: the current  registry
    :param namespace: the namespace of the model
    :param base: One of the base of the model
    :param properties: the properties of the model
    :rtype: new base
    """
    if hasattr(method, 'is_cache_method') and method.is_cache_method is True:
        if namespace not in registry.caches:
            registry.caches[namespace] = {attr: []}
        elif attr not in registry.caches[namespace]:
            registry.caches[namespace][attr] = []

        @lru_cache(maxsize=method.size)
        def wrapper(*args, **kwargs):
            return method(*args, **kwargs)

        wrapper.indentify = (namespace, attr)
        registry.caches[namespace][attr].append(wrapper)
        if method.is_cache_classmethod:
            return {attr: classmethod(wrapper)}
        else:
            return {attr: wrapper}

    return {}



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.blok import BlokManager
from sys import modules
from anyblok.imp import ImportManager, ImportManagerException
from os.path import join
from .mockblok import mockblok


tests_path = join('/'.join(__file__.split('/')[:-1]), 'mockblok')
with open(join(tests_path, 'mockfile.py'), 'r') as fp:
    initial_file = fp.read()


def python_version():
    import sys
    return '%d.%d' % (sys.version_info.major, sys.version_info.minor)


class TestImportManager(TestCase):

    def setUp(self):
        super(TestImportManager, self).setUp()
        BlokManager.bloks['mockblok'] = mockblok
        BlokManager.ordered_bloks.append('mockblok')

    def tearDown(self):
        super(TestImportManager, self).tearDown()
        if '.mockblok.' in modules:
            mod_2_del = []
            for mod in modules.keys():
                if '.mockblok.' in mod:
                    mod_2_del.append(mod)

            for mod in mod_2_del:
                del modules[mod]

        del BlokManager.bloks['mockblok']
        BlokManager.ordered_bloks.remove('mockblok')

    def test_has_blok(self):
        ImportManager.add('mockblok')
        self.assertEqual(ImportManager.has('mockblok'), True)
        self.assertEqual(ImportManager.has('mockblok2'), False)

    def test_get_blok(self):
        ImportManager.add('mockblok')
        ImportManager.get('mockblok')

    def test_get_unexisting_blok(self):
        try:
            ImportManager.get('mockblok2')
            self.fail('No watchdog for inexisting blok module')
        except ImportManagerException:
            pass

    def test_reload(self):
        blok = ImportManager.add('mockblok')
        blok.imports()
        from .mockblok.mockpackage import mockfile1, mockfile2
        from .mockblok.mockpackage import submockpackage
        from .mockblok import mockfile

        with open(join(tests_path, 'mockpackage', 'mockfile1.py'), 'w') as fp:
            fp.write("""foo = 'reload'""")

        with open(join(tests_path, 'mockpackage', 'submockpackage',
                       'mockfile2.py'), 'w') as fp:
            fp.write("""foo = 'reload'""")

        with open(join(tests_path, 'mockfile.py'), 'w') as fp:
            fp.write("""foo = 'reload'""")

        try:
            self.assertEqual(mockfile1.foo, 'bar')
            self.assertEqual(mockfile2.foo, 'bar')
            self.assertEqual(submockpackage.mockfile1.foo, 'bar')
            self.assertEqual(submockpackage.mockfile2.foo, 'bar')
            self.assertEqual(mockfile.foo, 'bar')
            blok.reload()
            self.assertEqual(mockfile1.foo, 'reload')
            self.assertEqual(mockfile2.foo, 'bar')
            self.assertEqual(submockpackage.mockfile1.foo, 'bar')
            self.assertEqual(submockpackage.mockfile2.foo, 'reload')
            self.assertEqual(mockfile.foo, 'reload')
        finally:
            with open(join(tests_path, 'mockfile.py'), 'w') as fp:
                fp.write(initial_file)

            with open(join(tests_path, 'mockpackage',
                           'mockfile1.py'), 'w') as fp:
                fp.write(initial_file)

            with open(join(tests_path, 'mockpackage', 'submockpackage',
                           'mockfile2.py'), 'w') as fp:
                fp.write(initial_file)

    def test_imports(self):
        blok = ImportManager.add('mockblok')
        blok.imports()
        from .mockblok.mockpackage import mockfile1, mockfile2
        from .mockblok.mockpackage import submockpackage
        from .mockblok import mockfile
        self.assertEqual(mockfile1.foo, 'bar')
        self.assertEqual(mockfile2.foo, 'bar')
        self.assertEqual(submockpackage.mockfile1.foo, 'bar')
        self.assertEqual(submockpackage.mockfile2.foo, 'bar')
        self.assertEqual(mockfile.foo, 'bar')



# flake8: noqa
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import reload_module_if_blok_is_reloading
from . import base
reload_module_if_blok_is_reloading(base)
from . import sqlbase
reload_module_if_blok_is_reloading(sqlbase)
from . import sqlviewbase
reload_module_if_blok_is_reloading(sqlviewbase)
from . import session
reload_module_if_blok_is_reloading(session)
from . import query
reload_module_if_blok_is_reloading(query)
from . import instrumentedlist
reload_module_if_blok_is_reloading(instrumentedlist)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import LargeBinary, Boolean, Integer


@Declarations.register(Declarations.Model.IO)
class Importer(Declarations.Mixin.IOMixin):

    file_to_import = LargeBinary(nullable=False)
    offset = Integer(default=0)
    nb_grouped_lines = Integer(nullable=False, default=50)
    commit_at_each_grouped = Boolean(default=True)
    check_import = Boolean(default=False)

    def run(self, blokname=None):
        return self.get_model(self.mode)(self, blokname=blokname).run()

    def get_key_mapping(self, key):
        Mapping = self.registry.IO.Mapping
        return Mapping.get(self.model, key)

    def commit(self):
        if self.check_import:
            return False
        elif not self.commit_at_each_grouped:
            return False

        self.registry.commit()
        return True

    def str2value(self, value, ctype, external_id=False, model=None):
        formater = self.get_formater(ctype)
        if external_id:
            return formater.externalIdStr2value(value, model)

        return formater.str2value(value, model)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok import release


class ModelBasedAuthorizationBlok(Blok):

    version = release.version
    author = 'Suzanne Jean-Sébastien'
    logo = '../anyblok-logo_alpha_256.png'

    @classmethod
    def import_declaration_module(cls):
        from . import models  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import models
        reload(models)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.environment import (EnvironmentManager,
                                 ThreadEnvironment,
                                 EnvironmentException)


class MockEnvironment:

    values = {}

    @classmethod
    def scoped_function_for_session(cls):
        return cls.__name__

    @classmethod
    def setter(cls, key, value):
        cls.values[key] = value

    @classmethod
    def getter(cls, key, default):
        return cls.values.get(key, default)


class TestEnvironment(TestCase):

    def setUp(self):
        super(TestEnvironment, self).setUp()
        EnvironmentManager.define_environment_cls(MockEnvironment)

    def tearDown(self):
        super(TestEnvironment, self).tearDown()
        EnvironmentManager.define_environment_cls(ThreadEnvironment)

    def test_set_and_get_variable(self):
        db_name = 'test db name'
        EnvironmentManager.set('db_name', db_name)
        self.assertEqual(EnvironmentManager.get('db_name'), db_name)

    def test_without_environment_for_set(self):
        # don't use define_environment_cls, because she must be verify
        EnvironmentManager.environment = None
        try:
            EnvironmentManager.set('db_name', 'test')
            self.fail('No watchdog for None environment')
        except EnvironmentException:
            pass

    def test_without_environment_for_get(self):
        # don't use define_environment_cls, because she must be verify
        EnvironmentManager.environment = None
        try:
            EnvironmentManager.get('db_name')
            self.fail('No watchdog for None environment')
        except EnvironmentException:
            pass

    def test_scoped_function_session(self):
        self.assertEqual(EnvironmentManager.scoped_function_for_session(),
                         MockEnvironment.scoped_function_for_session)

    def check_bad_define_environment(self, env):
        try:
            EnvironmentManager.define_environment_cls(env)
            self.fail("Bad environment class")
        except EnvironmentException:
            pass

    def test_bad_define_environment_scoped_function_session(self):

        class Env:

            @classmethod
            def setter(cls, key, value):
                pass

            @classmethod
            def getter(cls, key, default):
                pass

        self.check_bad_define_environment(Env)

        class Env(MockEnvironment):

            scoped_function_for_session = 'other'

        self.check_bad_define_environment(Env)

        class Env(MockEnvironment):

            def scoped_function_for_session(self):
                pass

        self.check_bad_define_environment(Env)

    def test_bad_define_environment_setter(self):

        class Env:

            @classmethod
            def scoped_function_for_session(cls):
                pass

            @classmethod
            def getter(cls, key, default):
                pass

        self.check_bad_define_environment(Env)

        class Env(MockEnvironment):

            setter = None

        self.check_bad_define_environment(Env)

        class Env(MockEnvironment):

            def setter(self, key, value):
                pass

        self.check_bad_define_environment(Env)

    def test_bad_define_environment_getter(self):

        class Env:

            @classmethod
            def scoped_function_for_session(cls):
                pass

            @classmethod
            def setter(cls, key, value):
                pass

        self.check_bad_define_environment(Env)

        class Env(MockEnvironment):

            getter = None

        self.check_bad_define_environment(Env)

        class Env(MockEnvironment):

            def getter(self, key, default):
                pass

        self.check_bad_define_environment(Env)


class TestThreadEnvironment(TestCase):

    def test_set_and_get_variable(self):
        db_name = 'test db name'
        EnvironmentManager.set('db_name', db_name)
        self.assertEqual(EnvironmentManager.get('db_name'), db_name)

    def test_scoped_function_session(self):
        self.assertEqual(EnvironmentManager.scoped_function_for_session(),
                         None)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from os.path import join
from os import walk
from logging import getLogger
import nose

from sqlalchemy import create_engine, event, MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.exc import (ProgrammingError, OperationalError,
                            InvalidRequestError)
from sqlalchemy_utils.functions import database_exists
from .config import Configuration, get_url
from .migration import Migration
from .blok import BlokManager
from .environment import EnvironmentManager
from .authorization.query import QUERY_WITH_NO_RESULTS, PostFilteredQuery
from anyblok.common import anyblok_column_prefix, naming_convention
from pkg_resources import iter_entry_points
from .logging import log
logger = getLogger(__name__)


class RegistryManagerException(Exception):
    """ Simple Exception for Registry """


class RegistryException(Exception):
    """ Simple Exception for Registry """


class RegistryConflictingException(Exception):
    """ Simple Exception for Registry """


def return_list(entry):
    if entry is not None and not isinstance(entry, (list, tuple)):
        entry = [entry]

    return entry


class RegistryManager:
    """ Manage the global registry

    Add new entry::

        RegistryManager.declare_entry('newEntry')
        RegistryManager.init_blok('newBlok')
        EnvironmentManager.set('current_blok', 'newBlok')
        RegistryManager.add_entry_in_register(
            'newEntry', 'oneKey', cls_)
        EnvironmentManager.set('current_blok', None)

    Remove an existing entry::

        if RegistryManager.has_entry_in_register('newBlok', 'newEntry',
                                                        'oneKey'):
            RegistryManager.remove_entry_in_register(
                'newBlok', 'newEntry', 'oneKey', cls_)

    get a new registry for a database::

        registry = RegistryManager.get('my database')

    """

    loaded_bloks = {}
    declared_entries = []
    declared_cores = []
    callback_pre_assemble_entries = {}
    callback_assemble_entries = {}
    callback_initialize_entries = {}
    callback_unload_entries = {}
    registries = {}
    needed_bloks = []

    @classmethod
    def has_blok(cls, blok):
        """ Return True if the blok is already loaded

        :param blok: name of the blok
        :rtype: boolean
        """

        return blok in cls.loaded_bloks

    @classmethod
    def clear(cls):
        """ Clear the registry dict to force the creation of new registry """
        registries = [r for r in cls.registries.values()]
        for registry in registries:
            registry.close()

    @classmethod
    def unload(cls):
        """Call all the unload callbacks"""
        for entry, unload_callback in cls.callback_unload_entries.items():
            logger.info('Unload: %r' % entry)
            unload_callback()

    @classmethod
    def get(cls, db_name, loadwithoutmigration=False, **kwargs):
        """ Return an existing Registry

        If the Registry doesn't exist then the Registry are created and added
        to registries dict

        :param db_name: the name of the database linked to this registry
        :rtype: ``Registry``
        """
        EnvironmentManager.set('db_name', db_name)
        if db_name in cls.registries:
            if loadwithoutmigration:
                logger.warning("loadwithoutmigration can not be used because "
                               "the registry for %r is already load" % db_name)
            return cls.registries[db_name]

        _Registry = Configuration.get('Registry', Registry)
        registry = _Registry(
            db_name, loadwithoutmigration=loadwithoutmigration, **kwargs)
        cls.registries[db_name] = registry
        return registry

    @classmethod
    def reload(cls):
        """ Reload the blok

        The purpose is to reload the python module to get changes in python
        file

        """
        for registry in cls.registries.values():
            registry.close_session()
            registry.Session = None
            registry.blok_list_is_loaded = False
            registry.reload()

    @classmethod
    def declare_core(cls, core):
        """ Add new core in the declared cores

        ::

            RegistryManager.declare_core('Core name')

            -----------------------------------------

            @Declarations.register(Declarations.Core)
            class ``Core name``:
                ...

        .. warning::

            The core must be declared in the application, not in the bloks
            The declaration must be done before the loading of the bloks

        :param core: core name
        """

        if core not in cls.declared_cores:
            cls.declared_cores.append(core)

    @classmethod
    def undeclare_core(cls, core):
        if core in cls.declared_cores:
            cls.declared_cores.remove(core)

    @classmethod
    def declare_entry(cls, entry,
                      pre_assemble_callback=None,
                      assemble_callback=None,
                      initialize_callback=None):
        """ Add new entry in the declared entries

        ::

            def assemble_callback(registry):
                ...

            def initialize_callback(registry):
                ...

            RegistryManager.declare_entry(
                'Entry name', assemble_callback=assemble_callback,
                initialize_callback=initialize_callback)

            @Declarations.register(Declarations.``Entry name``)
            class MyClass:
                ...

        .. warning::

            The entry must be declared in the application, not in the bloks
            The declaration must be done before the loading of the bloks

        :param entry: entry name
        :param assemble_callback: function callback to call to assemble
        :param initialize_callback: function callback to call to init after
            assembling
        """

        if entry not in cls.declared_entries:
            cls.declared_entries.append(entry)

            if pre_assemble_callback:
                cls.callback_pre_assemble_entries[entry] = pre_assemble_callback

            if assemble_callback:
                cls.callback_assemble_entries[entry] = assemble_callback

            if initialize_callback:
                cls.callback_initialize_entries[entry] = initialize_callback

    @classmethod
    def declare_unload_callback(cls, entry, unload_callback):
        """Save a unload callback in registry Manager

        :param entry: declaration type name
        :param unload_callback: classmethod pointer
        """
        cls.callback_unload_entries[entry] = unload_callback

    @classmethod
    def undeclare_entry(cls, entry):
        if entry in cls.declared_entries:
            cls.declared_entries.remove(entry)

            if entry in cls.callback_pre_assemble_entries:
                del cls.callback_pre_assemble_entries[entry]

            if entry in cls.callback_assemble_entries:
                del cls.callback_assemble_entries[entry]

            if entry in cls.callback_initialize_entries:
                del cls.callback_initialize_entries[entry]

    @classmethod
    def init_blok(cls, blokname):
        """ init one blok to be known by the RegistryManager

        All bloks loaded must be initialized because the registry will be
        created with this information

        :param blokname: name of the blok
        """
        blok = {
            'Core': {core: [] for core in cls.declared_cores},
            'properties': {},
            'removed': [],
        }
        for de in cls.declared_entries:
            blok[de] = {'registry_names': []}

        cls.loaded_bloks[blokname] = blok

    @classmethod
    def has_core_in_register(cls, blok, core):
        """ Return True if One Class exist in this blok for this core

        :param blok: name of the blok
        :param core: is the existing core name
        """
        return len(cls.loaded_bloks[blok]['Core'][core]) > 0

    @classmethod
    def add_core_in_register(cls, core, cls_):
        """ Load core in blok

        warning the global var current_blok must be filled on the good blok

        :param core: is the existing core name
        :param ``cls_``: Class of the Core to save in loaded blok target
            registry
        """

        current_blok = EnvironmentManager.get('current_blok')
        cls.loaded_bloks[current_blok]['Core'][core].append(cls_)

    @classmethod
    def remove_in_register(cls, cls_):
        """ Remove Class in blok and in entry

        :param ``cls_``: Class of the entry / key to remove in loaded blok
        """
        current_blok = EnvironmentManager.get('current_blok')
        removed = cls.loaded_bloks[current_blok]['removed']
        if cls_ not in removed:
            removed.append(cls_)

    @classmethod
    def has_entry_in_register(cls, blok, entry, key):
        """ Return True if One Class exist in this blok for this entry

        :param blok: name of the blok
        :param entry: is the existing entry name
        :param key: is the existing key in the entry
        """
        if entry not in cls.loaded_bloks[blok]:
            return False

        if key not in cls.loaded_bloks[blok][entry]:
            return False

        return len(cls.loaded_bloks[blok][entry][key]['bases']) > 0

    @classmethod
    def add_entry_in_register(cls, entry, key, cls_, **kwargs):
        """ Load entry in blok

        warning the global var current_blok must be filled on the good blok
        :param entry: is the existing entry name
        :param key: is the existing key in the entry
        :param ``cls_``: Class of the entry / key to remove in loaded blok
        """

        bases = []

        for base in cls_.__bases__:
            if hasattr(base, '__registry_name__'):
                bases.append(base)

        setattr(cls_, '__anyblok_bases__', bases)

        cb = EnvironmentManager.get('current_blok')

        if key not in cls.loaded_bloks[cb][entry]:
            cls.loaded_bloks[cb][entry][key] = {
                'bases': [],
                'properties': {},
            }

        cls.loaded_bloks[cb][entry][key]['properties'].update(kwargs)
        # Add before in registry because it is the same order than the
        # inheritance __bases__ and __mro__
        cls.loaded_bloks[cb][entry][key]['bases'].insert(0, cls_)

        if key not in cls.loaded_bloks[cb][entry]['registry_names']:
            cls.loaded_bloks[cb][entry]['registry_names'].append(key)

    @classmethod
    def get_entry_properties_in_register(cls, entry, key):
        cb = EnvironmentManager.get('current_blok')
        if key not in cls.loaded_bloks[cb][entry]:
            return {}

        return cls.loaded_bloks[cb][entry][key]['properties'].copy()

    @classmethod
    def has_blok_property(cls, property_):
        """ Return True if the property exists in blok

        :param property\_: name of the property
        """
        blok = EnvironmentManager.get('current_blok')

        if property_ in cls.loaded_bloks[blok]['properties']:
            return True

        return False

    @classmethod
    def add_or_replace_blok_property(cls, property_, value):
        """ Save the value in the properties

        :param property\_: name of the property
        :param value: the value to save, the type is not important
        """
        blok = EnvironmentManager.get('current_blok')
        cls.loaded_bloks[blok]['properties'][property_] = value

    @classmethod
    def get_blok_property(cls, property_, default=None):
        """ Return the value in the properties

        :param property\_: name of the property
        :param default: return default If not entry in the property
        """
        blok = EnvironmentManager.get('current_blok')
        return cls.loaded_bloks[blok]['properties'].get(property_, default)

    @classmethod
    def remove_blok_property(cls, property_):
        """ Remove the property if exist

        :param property\_: name of the property
        """
        blok = EnvironmentManager.get('current_blok')
        if cls.has_blok_property(property_):
            del cls.loaded_bloks[blok]['properties'][property_]

    @classmethod
    def add_needed_bloks(cls, *bloks):
        for blok in bloks:
            if blok not in cls.needed_bloks:
                cls.needed_bloks.append(blok)

    @classmethod
    def get_needed_bloks(cls):
        return cls.needed_bloks


class Registry:
    """ Define one registry

    A registry is linked to a database, and stores the definition of the
    installed Bloks, Models, Mixins for this database::

        registry = Registry('My database')
    """

    def __init__(self, db_name, loadwithoutmigration=False, unittest=False,
                 **kwargs):
        self.db_name = db_name
        self.loadwithoutmigration = loadwithoutmigration
        self.unittest = unittest
        self.additional_setting = kwargs
        self.init_engine(db_name=db_name)
        self.init_bind()
        self.registry_base = type("RegistryBase", tuple(), {
            'registry': self,
            'Env': EnvironmentManager})
        self.withoutautomigration = Configuration.get('withoutautomigration')
        self.ini_var()
        self.Session = None
        self.nb_query_bases = self.nb_session_bases = 0
        self.blok_list_is_loaded = False
        self.pre_assemble_entries()
        self.load()

    def init_bind(self):
        """Initialize the bind"""
        if self.unittest:
            self.bind = self.engine.connect()
            self.unittest_transaction = self.bind.begin()
        else:
            self.bind = self.engine
            self.unittest_transaction = None

    def init_engine_options(self):
        """Define the options to initialize the engine"""
        return dict(
            echo=Configuration.get('db_echo') or False,
            max_overflow=Configuration.get('db_max_overflow') or 10,
            echo_pool=Configuration.get('db_echo_pool') or False,
            pool_size=Configuration.get('db_pool_size') or 5,
            isolation_level=self.additional_setting.get(
                'isolation_level',
                Configuration.get('isolation_level', 'READ_UNCOMMITTED')
            ),
        )

    def init_engine(self, db_name=None):
        """Define the engine

        :param db_name: name of the database to link
        """
        kwargs = self.init_engine_options()
        url = Configuration.get('get_url', get_url)(db_name=db_name)
        self.rw_engine = create_engine(url, **kwargs)

    @property
    def engine(self):
        """property to get the engine"""
        return self.rw_engine

    def ini_var(self):
        """ Initialize the var to load the registry """
        self.loaded_namespaces = {}
        self.declarativebase = None
        self.loaded_bloks = {}
        self.loaded_registries = {x + '_names': []
                                  for x in RegistryManager.declared_entries}
        self.loaded_cores = {core: []
                             for core in RegistryManager.declared_cores}
        self.ordered_loaded_bloks = []
        self.loaded_namespaces = {}
        self.children_namespaces = {}
        self.properties = {}
        self.removed = []
        EnvironmentManager.set('_precommit_hook', [])
        EnvironmentManager.set('_postcommit_hook', [])
        self._sqlalchemy_known_events = []
        self.expire_attributes = {}

    @classmethod
    def db_exists(cls, db_name=None):
        if not db_name:
            raise RegistryException('db_name is required')

        url = Configuration.get('get_url', get_url)(db_name=db_name)
        return database_exists(url)

    def listen_sqlalchemy_known_event(self):
        for e, namespace, method in self._sqlalchemy_known_events:
            if hasattr(method, 'get_attribute'):
                method = method.get_attribute(self)

            event.listen(e.mapper(self, namespace, usehybrid=False), e.event,
                         method, *e.args, **e.kwargs)

    def remove_sqlalchemy_known_event(self):
        for e, namespace, method in self._sqlalchemy_known_events:
            try:
                if hasattr(method, 'get_attribute'):
                    method = method.get_attribute(self)

                event.remove(e.mapper(self, namespace), e.event,
                             method)
            except InvalidRequestError:
                pass

    def get(self, namespace):
        """ Return the namespace Class

        :param namespace: namespace to get from the registry str
        :rtype: namespace cls
        :exception: RegistryManagerException
        """
        if namespace not in self.loaded_namespaces:
            raise RegistryManagerException(
                "No namespace %r loaded" % namespace)

        return self.loaded_namespaces[namespace]

    def has(self, namespace):
        return True if namespace in self.loaded_namespaces else False

    def get_bloks_by_states(self, *states):
        """ Return the bloks in these states

        :param states: list of the states
        :rtype: list of blok's name
        """
        if not states:
            return []

        res = []
        query = """
            SELECT system_blok.name
            FROM system_blok
            WHERE system_blok.state in ('%s')
            ORDER BY system_blok.order""" % "', '".join(states)
        try:
            res = self.execute(query).fetchall()
        except (ProgrammingError, OperationalError):
            pass

        if res:
            return [x[0] for x in res]

        return []

    def get_bloks_to_load(self):
        """ Return the bloks to load by the registry

        :rtype: list of blok's name
        """
        return self.get_bloks_by_states('installed', 'toupdate')

    def get_bloks_to_install(self, loaded):
        """ Return the bloks to install in the registry

        :rtype: list of blok's name
        """
        toinstall = self.get_bloks_by_states('toinstall')
        for blok in BlokManager.auto_install:
            if blok not in (toinstall + loaded):
                toinstall.append(blok)

        for blok in RegistryManager.get_needed_bloks():
            if blok not in (toinstall + loaded):
                toinstall.append(blok)

        if toinstall and self.withoutautomigration:
            raise RegistryManagerException(
                "Install modules %r is forbidden with no auto migration "
                "mode" % toinstall)

        return toinstall

    def check_permission(self, target, principals, permission):
        """Check that one of the principals has permisson on target.

        :param target: model instance (record) or class. Checking a permission
                       on a model class with a policy that needs to work on
                       records is considered a configuration error: the policy
                       has the right to fail.
        :param principals: list, set or tuple of strings
        :rtype: bool
        """
        return self.lookup_policy(target, permission).check(
            target, principals, permission)

    def wrap_query_permission(self, query, principals, permission, models=()):
        """Wrap query to return only authorized results

        :param principals: list, set or tuple of strings
        :param models: models on which to apply security filtering. If
                       not supplied, it will be infered from the query. The
                       length and ordering much match that of expected results.
        :returns: a query-like object, implementing the results fetching API,
                  but that can't be further filtered.

        This method calls all the relevant policies to apply pre- and
        post-filtering. Although postfiltering is discouraged in authorization
        policies for performance and expressiveness (limit, offset),
        there are cases for which it is unavoidable, or in which the tradeoff
        goes the other way.

        In normal operation, the relevant models are infered directly from
        the query.
        For join situations, and more complex queries, the caller has control
        on the models on which to exert permission checking.

        For instance, it might make sense to use a join between Model1 and
        Model2 to actually constrain Model1 (on which permission filtering
        should occur) by information contained in Model2, even if the passed
        principals should not grant access to the relevant Model2 records.
        """
        if not models:
            models = []
            for column in query.column_descriptions:
                if column['aliased']:
                    # actually, think aliases could work almost direcly
                    # it's just a matter of documenting that what the policy
                    # gets may be an alias instead of a model.
                    raise NotImplementedError(
                        "Sorry, table/model aliases aren't supported yet. "
                        "Here's the unsupported column: %r" % column)
                if not issubclass(column['type'], self.registry_base):
                    raise NotImplementedError(
                        "Sorry, only model columns are supported for now. "
                        "Here is the unsupported one: %r" % column)
                models.append(column['type'])

        postfilters = {}
        for model in models:
            policy = self.lookup_policy(model, permission)
            query = policy.filter(model, query, principals, permission)
            if query is False:  # TODO use a dedicated singleton ?
                return QUERY_WITH_NO_RESULTS
            if policy.postfilter is not None:
                postfilters[model] = lambda rec: policy.postfilter(
                    rec, principals, permission)
        return PostFilteredQuery(query, postfilters)

    def lookup_policy(self, target, permission):
        """Return the policy instance that applies to target or its model.

        :param target: model class or instance

        If a policy is declared for the precise permission, it is returned.
        Otherwise, the default policy for that model is returned.
        By ultimate default the special
        :class:`anyblok.authorization.rule.DenyAll` is returned.
        """
        model_name = target.__registry_name__
        policy = self._authz_policies.get((model_name, permission))
        if policy is not None:
            return policy

        policy = self._authz_policies.get(model_name)
        if policy is not None:
            return policy

        return self._authz_policies.get(None)

    def load_entry(self, blok, entry):
        """ load one entry type for one blok

        :param blok: name of the blok
        :param entry: declaration type to load
        """
        _entry = RegistryManager.loaded_bloks[blok][entry]
        for key in _entry['registry_names']:
            v = _entry[key]
            if key not in self.loaded_registries:
                self.loaded_registries[key] = {'properties': {}, 'bases': []}

            self.loaded_registries[key]['properties'].update(v['properties'])
            old_bases = [] + self.loaded_registries[key]['bases']
            self.loaded_registries[key]['bases'] = v['bases']
            self.loaded_registries[key]['bases'] += old_bases
            self.loaded_registries[entry + '_names'].append(key)

    def load_core(self, blok, core):
        """ load one core type for one blok

        :param blok: name of the blok
        :param core: the core name to load
        """
        if core in RegistryManager.loaded_bloks[blok]['Core']:
            bases = RegistryManager.loaded_bloks[blok]['Core'][core]
            for base in bases:
                self.loaded_cores[core].insert(0, base)
        else:
            logger.warning('No Core %r found' % core)

    def load_properties(self, blok):
        properties = RegistryManager.loaded_bloks[blok]['properties']
        for k, v in properties.items():
            if k not in self.properties:
                self.properties[k] = v
            elif isinstance(self.properties[k], dict) and isinstance(v, dict):
                self.properties.update(v)
            elif isinstance(self.properties[k], list) and isinstance(v, list):
                self.properties.extend(v)
            else:
                self.properties[k] = v

    def load_removed(self, blok):
        for removed in RegistryManager.loaded_bloks[blok]['removed']:
            if removed not in self.removed:
                self.removed.append(removed)

    def load_bloks(self, bloks, toinstall, toload, required=True):
        for blok in bloks:
            if required:
                if not self.load_blok(blok, toinstall, toload):
                    raise RegistryManagerException(
                        "Required blok %r not found" % blok)
            elif toinstall or blok in toload:
                self.load_blok(blok, toinstall, toload)

    def load_blok(self, blok, toinstall, toload):
        """ load on blok, load all the core and all the entry for one blok

        :param blok: name of the blok
        :exception: RegistryManagerException
        """
        if blok in self.ordered_loaded_bloks:
            return True

        if blok not in BlokManager.bloks:
            return False

        b = BlokManager.bloks[blok](self)
        self.load_bloks(b.required + b.conditional, toinstall, toload)
        self.load_bloks(b.optional, toinstall, toload, required=False)

        for core in RegistryManager.declared_cores:
            self.load_core(blok, core)

        for entry in RegistryManager.declared_entries:
            self.load_entry(blok, entry)

        self.load_properties(blok)
        self.load_removed(blok)
        self.loaded_bloks[blok] = b
        self.ordered_loaded_bloks.append(blok)
        logger.debug("Blok %r loaded" % blok)
        return True

    def check_dependencies(self, blok, dependencies_to_install, toinstall):
        if blok in dependencies_to_install:
            return True

        if blok not in BlokManager.bloks:
            return False

        b = BlokManager.bloks[blok](self)
        for required in b.required:
            if not self.check_dependencies(required, dependencies_to_install,
                                           toinstall):
                raise RegistryManagerException(
                    "%r: Required blok not found %r" % (blok, required))

        for optional in b.optional:
            self.check_dependencies(
                optional, dependencies_to_install, toinstall)

        if blok not in toinstall:
            dependencies_to_install.append(blok)

        return True

    def update_to_install_blok_dependencies_state(self, toinstall):
        dependencies_to_install = []

        for blok in toinstall:
            self.check_dependencies(blok, dependencies_to_install, toinstall)

        if dependencies_to_install:
            query = """
                update system_blok
                set state='toinstall'
                where name in ('%s')
                and state = 'uninstalled'""" % "', '".join(
                dependencies_to_install)
            try:
                self.execute(query)
            except (ProgrammingError, OperationalError):
                pass

            return True

        return False

    def execute(self, *args, **kwargs):
        if self.Session:
            return self.session.execute(*args, **kwargs)
        else:
            conn = None
            try:
                conn = self.engine.connect()
                return conn.execute(*args, **kwargs)
            finally:
                if conn:
                    conn.close()

    def get_namespace(self, parent, child):
        if hasattr(parent, child) and getattr(parent, child):
            return getattr(parent, child)

        tmpns = type(child, tuple(), {'children_namespaces': {}})
        if hasattr(parent, 'children_namespaces'):
            parent.children_namespaces[child] = tmpns

        setattr(parent, child, tmpns)
        return tmpns

    def final_namespace(self, parent, child, base):
        if hasattr(parent, child) and getattr(parent, child):
            other_base = self.get_namespace(parent, child)
            other_base = other_base.children_namespaces.copy()
            for ns, cns in other_base.items():
                setattr(base, ns, cns)

            if hasattr(parent, 'children_namespaces'):
                if child in parent.children_namespaces:
                    parent.children_namespaces[child] = base

        elif hasattr(parent, 'children_namespaces'):
            parent.children_namespaces[child] = base

        setattr(parent, child, base)

    def add_in_registry(self, namespace, base):
        """ Add a class as an attribute of the registry

        :param namespace: tree path of the attribute
        :param base: class to add
        """
        namespace = namespace.split('.')[1:]

        def update_namespaces(parent, namespaces):
            if len(namespaces) == 1:
                self.final_namespace(parent, namespaces[0], base)
            else:
                new_parent = self.get_namespace(parent, namespaces[0])
                update_namespaces(new_parent, namespaces[1:])

        update_namespaces(self, namespace)

    def create_session_factory(self):
        """Create the SQLA Session factory

        in function of the Core Session class ans the Core Qery class
        """
        if self.Session is None or self.must_recreate_session_factory():
            bind = self.bind
            if self.Session:
                if not self.withoutautomigration:
                    # this is the only case to use commit in the construction
                    # of the registry
                    self.commit()

                # remove all existing instance to create a new instance
                # because the instance are cached
                self.Session.remove()

            query_bases = [] + self.loaded_cores['Query']
            query_bases += [self.registry_base]
            Query = type('Query', tuple(query_bases), {})
            session_bases = [self.registry_base] + self.loaded_cores['Session']
            Session = type('Session', tuple(session_bases), {
                'registry_query': Query})

            extension = self.additional_setting.get('sa.session.extension')
            if extension:
                extension = extension()
            self.Session = scoped_session(
                sessionmaker(bind=bind, class_=Session, extension=extension),
                EnvironmentManager.scoped_function_for_session())
            self.nb_query_bases = len(self.loaded_cores['Query'])
            self.nb_session_bases = len(self.loaded_cores['Session'])
            self.apply_session_events()
        else:
            self.flush()

    def must_recreate_session_factory(self):
        """Check if the SQLA Session Factory must be destroy and recreate

        :rtype: Boolean, True if nb Core Session/Query inheritance change
        """
        nb_query_bases = len(self.loaded_cores['Query'])
        if nb_query_bases != self.nb_query_bases:
            return True

        nb_session_bases = len(self.loaded_cores['Session'])
        if nb_session_bases != self.nb_session_bases:
            return True

        return False

    @log(logger, level='debug')
    def load(self):
        """ Load all the namespaces of the registry

        Create all the table, make the shema migration
        Update Blok, Model, Column rows
        """
        mustreload = False
        blok2install = None
        try:
            self.declarativebase = declarative_base(
                metadata=MetaData(naming_convention=naming_convention),
                class_registry=dict(registry=self))
            toload = self.get_bloks_to_load()
            toinstall = self.get_bloks_to_install(toload)
            if self.update_to_install_blok_dependencies_state(toinstall):
                toinstall = self.get_bloks_to_install(toload)
            if self.loadwithoutmigration and not toload and toinstall:
                logger.warning("Impossible to use loadwithoumigration")
                self.loadwithoutmigration = False

            self.load_bloks(toload, False, toload)
            if toinstall and not self.loadwithoutmigration:
                blok2install = toinstall[0]
                self.load_blok(blok2install, True, toload)

            instrumentedlist_base = [] + self.loaded_cores['InstrumentedList']
            instrumentedlist_base += [list]
            self.InstrumentedList = type(
                'InstrumentedList', tuple(instrumentedlist_base), {})
            self.assemble_entries()
            self.create_session_factory()

            mustreload = self.apply_model_schema_on_table(
                blok2install) or mustreload

        except Exception as e:
            self.close()
            raise e

        test_blok = blok2install and Configuration.get(
            'test_blok_at_install')
        selected_bloks = return_list(Configuration.get('selected_bloks'))
        in_selected_bloks = blok2install in (selected_bloks or [blok2install])
        unwanted_bloks = return_list(Configuration.get('unwanted_bloks'))
        not_in_unwanted_bloks = blok2install not in (unwanted_bloks or [])

        if test_blok and in_selected_bloks and not_in_unwanted_bloks:
            self.System.Blok.load_all()
            self.run_test(blok2install)
            if len(toinstall) > 1 or mustreload:
                self.reload()

        else:
            if len(toinstall) > 1 or mustreload:
                self.reload()
            else:
                self.System.Blok.load_all()

        self.loadwithoutmigration = False

    def apply_session_events(self):
        """Add session events

        the session event come from:

        * entrypoints: ``anyblok.session.event``
        * registry additional_setting: ``anyblok.session.event``
        """
        for i in iter_entry_points('anyblok.session.event'):
            logger.info('Update session event from entrypoint %r' % i)
            i.load()(self.session)

        for funct in self.additional_setting.get('anyblok.session.event', []):
            logger.info('Update session event %r' % funct)
            funct(self.session)

    def run_test(registry, blok2install):
        startpath = BlokManager.getPath(blok2install)

        class ContextSuite(nose.suite.ContextSuite):

            def __init__(self, *args, **kwargs):
                super(ContextSuite, self).__init__(*args, **kwargs)
                if self.context is not None:
                    self.context.registry = registry

        class ContextSuiteFactory(nose.suite.ContextSuiteFactory):
            suiteClass = ContextSuite

        class TestLoader(nose.loader.TestLoader):

            def __init__(self, *args, **kwargs):
                super(TestLoader, self).__init__(*args, **kwargs)
                self.suiteClass = ContextSuiteFactory(config=self.config)

        for root, dirs, _ in walk(startpath):
            if 'tests' in dirs:
                nose.run(defaultTest=[join(root, 'tests')],
                         testLoader=TestLoader(),
                         argv=['-v', '-s'])

    def assemble_entries(self):
        for entry in RegistryManager.declared_entries:
            if entry in RegistryManager.callback_assemble_entries:
                logger.debug('Assemble %r entry' % entry)
                RegistryManager.callback_assemble_entries[entry](self)

    def pre_assemble_entries(self):
        for entry in RegistryManager.declared_entries:
            if entry in RegistryManager.callback_pre_assemble_entries:
                logger.debug('Pre assemble %r entry' % entry)
                RegistryManager.callback_pre_assemble_entries[entry](self)

    def apply_model_schema_on_table(self, blok2install):
        # replace the engine by the session.connection for bind attribute
        # because session.connection is already the connection use
        # by blok, migration and all write on the data base
        # or use engine for bind, force create_all method to create new
        # new connection, this new connection have not acknowedge of the
        # data in the session.connection, and risk of bad lock on the
        # tables
        if self.loadwithoutmigration:
            return

        if not self.withoutautomigration:
            self.declarativebase.metadata.create_all(self.connection())

        self.migration = Configuration.get('Migration', Migration)(self)
        query = """
            SELECT name, installed_version
            FROM system_blok
            WHERE
                (state = 'toinstall' AND name = '%s')
                OR state = 'toupdate'""" % blok2install
        res = self.execute(query).fetchall()
        if res:
            for blok, installed_version in res:
                b = BlokManager.get(blok)(self)
                b.pre_migration(installed_version)

            self.migration.auto_upgrade_database()
            for blok, installed_version in res:
                b = BlokManager.get(blok)(self)
                b.post_migration(installed_version)

        else:
            self.migration.auto_upgrade_database()

        self.listen_sqlalchemy_known_event()
        mustreload = False
        for entry in RegistryManager.declared_entries:
            if entry in RegistryManager.callback_initialize_entries:
                logger.debug('Initialize %r entry' % entry)
                r = RegistryManager.callback_initialize_entries[entry](
                    self)
                mustreload = mustreload or r

        return mustreload

    def expire(self, obj, attribute_names=None):
        """Expire object in session, you can define some attribute which are
        expired::

            registry.expire(instance, ['attr1', 'attr2', ...])

        :param obj: instance of ``Model``
        :param attribute_names: list of string, names of the attr to expire
        """
        if attribute_names:
            hybrid_property_columns = (
                obj.__class__.get_hybrid_property_columns()
            )
            attribute_names = [
                (anyblok_column_prefix + x
                 if x in hybrid_property_columns else x)
                for x in attribute_names
            ]

        self.session.expire(obj, attribute_names=attribute_names)

    def expire_all(self):
        """Expire all the objects in session

        ::
            registry.expire_all()

        """
        self.session.expire_all()

    def expunge(self, obj):
        """Expunge instance of the session, remove all links of this instance
        in the session::

            registry.expunge(instance_of_model)

        """
        self.session.expunge(obj)

    def refresh(self, obj, attribute_names=None):
        """Expire  and reload object in session, you can define some attribute
        which are refreshed::

            registry.refresh(instance, ['attr1', 'attr2', ...])

        :param obj: instance of ``Model``
        :param attribute_names: list of string, names of the attr to refresh
        """
        if attribute_names:
            attribute_names = [
                (anyblok_column_prefix + x
                 if x in obj.hybrid_property_columns else x)
                for x in attribute_names]

        self.session.refresh(obj, attribute_names=attribute_names)

    def rollback(self, *args, **kwargs):
        logger.debug('[ROLLACK] with args=%r and kwargs = %r', args, kwargs)
        self.session.rollback(*args, **kwargs)
        EnvironmentManager.set('_precommit_hook', [])
        EnvironmentManager.set('_postcommit_hook', [])

    def close_session(self):
        """ Close only the session, not the registry
        After the call of this method the registry won't be usable
        you should use close method which call this method
        """

        if self.unittest_transaction:
            self.unittest_transaction.rollback()

        if self.Session:
            session = self.Session()
            session.rollback()
            session.expunge_all()
            session.close_all()

        if self.unittest_transaction:
            self.unittest_transaction.close()
            self.bind.close()

    def close(self):
        """Release the session, connection and engine"""
        self.close_session()
        self.engine.dispose()
        if self.db_name in RegistryManager.registries:
            del RegistryManager.registries[self.db_name]

    def __getattr__(self, attribute):
        # TODO safe the call of session for reload
        if self.Session:
            session = self.Session()
            if attribute == 'session':
                return session
            if hasattr(session, attribute):
                return getattr(session, attribute)

        else:
            return super(Registry, self).__getattr__(attribute)

    def precommit_hook(self, registryname, method, *args, **kwargs):
        """ Add a method in the precommit_hook list

        a precommit hook is a method called just before the commit, it is used
        to call this method once, because a hook is saved only once

        :param registryname: namespace of the model
        :param method: method to call on the registryname
        :param put_at_the_end_if_exist: if true and hook allready exist then the
            hook are moved at the end
        """
        put_at_the_end_if_exist = kwargs.pop('put_at_the_end_if_exist', False)

        entry = (registryname, method, args, kwargs)
        _precommit_hook = EnvironmentManager.get('_precommit_hook', [])
        if entry in _precommit_hook:
            if put_at_the_end_if_exist:
                _precommit_hook.remove(entry)
                _precommit_hook.append(entry)

        else:
            _precommit_hook.append(entry)

        EnvironmentManager.set('_precommit_hook', _precommit_hook)

    def postcommit_hook(self, registryname, method, *args, **kwargs):
        """ Add a method in the postcommit_hook list

        a precommit hook is a method called just after the commit, it is used
        to call this method once, because a hook is saved only once

        you can choice if the hook is called in function of ``call_only_if``:

        * ``commited``: Call if the commit is done without exception
        * ``raised``: Call if one exception was raised
        * ``always``: Always call

        :param registryname: namespace of the model
        :param method: method to call on the registryname
        :param put_at_the_end_if_exist: if true and hook allready exist then the
            hook are moved at the end
        :param call_only_if: ['commited' (default), 'raised', 'always']
        """
        put_at_the_end_if_exist = kwargs.pop('put_at_the_end_if_exist', False)
        call_only_if = kwargs.pop('call_only_if', 'commited')

        entry = (registryname, method, call_only_if, args, kwargs)
        _postcommit_hook = EnvironmentManager.get('_postcommit_hook', [])
        if entry in _postcommit_hook:
            if put_at_the_end_if_exist:
                _postcommit_hook.remove(entry)
                _postcommit_hook.append(entry)

        else:
            _postcommit_hook.append(entry)

        EnvironmentManager.set('_postcommit_hook', _postcommit_hook)

    def apply_precommit_hook(self):
        hooks = []
        _precommit_hook = EnvironmentManager.get('_precommit_hook')
        if _precommit_hook:
            hooks.extend(_precommit_hook)

        for hook in hooks:
            Model = self.loaded_namespaces[hook[0]]
            method = hook[1]
            a = hook[2]
            kw = hook[3]
            getattr(Model, method)(*a, **kw)
            _precommit_hook.remove(hook)

    def apply_postcommit_hook(self, withexception=False):
        hooks = []
        _postcommit_hook = EnvironmentManager.get('_postcommit_hook')
        if _postcommit_hook:
            hooks.extend(_postcommit_hook)

        for hook in hooks:
            registryname, method, call_only_if, a, kw = hook
            if withexception is False and call_only_if == 'raised':
                _postcommit_hook.remove(hook)
                continue

            if withexception is True and call_only_if == 'commited':
                _postcommit_hook.remove(hook)
                continue

            Model = self.loaded_namespaces[registryname]
            try:
                getattr(Model, method)(*a, **kw)
            except Exception as e:
                logger.exception(str(e))
            finally:
                _postcommit_hook.remove(hook)

    def commit(self, *args, **kwargs):
        """ Overload the commit method of the SqlAlchemy session """
        logger.debug('[COMMIT] with args=%r and kwargs = %r', args, kwargs)
        try:
            self.apply_precommit_hook()
            self.session_commit(*args, **kwargs)
            try:
                self.apply_postcommit_hook(withexception=False)
            except Exception as e:
                logger.exception(str(e))
        except Exception as e:
            try:
                self.apply_postcommit_hook(withexception=True)
            except Exception as e:
                logger.exception(str(e))

            raise e

    def flush(self):
        if not self.session._flushing:
            self.session.flush()

    def session_commit(self, *args, **kwargs):
        if self.Session:
            session = self.Session()
            session.commit(*args, **kwargs)

    def clean_model(self):
        """ Clean the registry of all the namespaces """
        for model in self.loaded_namespaces.keys():
            name = model.split('.')[1]
            if hasattr(self, name) and getattr(self, name):
                setattr(self, name, None)

    @log(logger, level='debug')
    def complete_reload(self):
        """ Reload the code and registry"""
        BlokManager.reload()
        RegistryManager.reload()

    @log(logger, level='debug')
    def reload(self):
        """ Reload the registry, close session, clean registry, reinit var """
        # self.close_session()
        self.remove_sqlalchemy_known_event()
        self.clean_model()
        self.ini_var()
        self.load()

    def get_bloks(self, blok, filter_states, filter_modes):
        Blok = self.System.Blok
        definition_blok = BlokManager.bloks[blok]
        bloks_name = []
        for filter_mode in filter_modes:
            bloks_name.extend(getattr(definition_blok, filter_mode))

        if not bloks_name:
            return []

        query = Blok.query()
        query = query.filter(Blok.name.in_(bloks_name))
        query = query.filter(Blok.state.in_(filter_states))
        return query.all().name

    def check_conflict_with(self, blok):
        Blok = self.System.Blok
        definition_blok = BlokManager.bloks[blok]
        conflicting_bloks = []
        conflicting_bloks.extend(definition_blok.conflicting)
        conflicting_bloks.extend(definition_blok.conflicting_by)

        query = Blok.query()
        query = query.filter(Blok.name.in_(conflicting_bloks))
        query = query.filter(
            Blok.state.in_(['installed', 'toinstall', 'toupdate']))
        if query.count():
            raise RegistryConflictingException(
                "Installation of the blok %r is forbidden, because the blok "
                "%r conflict with the blok(s) : %r" % (
                    blok, blok, [str(x) for x in query.all()]))

    def apply_state(self, blok_name, state, in_states):
        """Apply the state of the blok name

        :param blok_name: the name of the blok
        :param state: the state to apply
        :param in_states: the blok must be in this state
        :exception: RegistryException
        """
        Blok = self.System.Blok
        query = Blok.query().filter(Blok.name == blok_name)
        blok = query.first()
        if blok is None:
            raise RegistryException(
                "Blok %r not found in entry point declarations" %
                blok_name)

        if blok.state == state:
            logger.debug("Does not change state for blok %s because is the "
                         "same %s" % (blok_name, state))
            return

        if blok.state not in in_states:
            raise RegistryException(
                "Apply state %r is forbidden because the state %r of "
                "blok %r is not one of %r" % (
                    state, blok.state, blok_name, in_states))

        logger.info("Change state %s => %s for blok %s" % (
            blok.state, state, blok_name))
        query.update({Blok.state: state})

    @log(logger, level='debug', withargs=True)
    def upgrade(self, install=None, update=None, uninstall=None):
        """ Upgrade the current registry

        :param install: list of the blok to install
        :param update: list of the blok to update
        :param uninstall: list of the blok to uninstall
        :exception: RegistryException
        """
        Blok = self.System.Blok

        def upgrade_state_bloks(state):
            def wrap(bloks):
                for blok in bloks:
                    if state == 'toinstall':
                        self.check_conflict_with(blok)
                        self.apply_state(blok, state, ['uninstalled'])
                        upgrade_state_bloks(state)(
                            self.get_bloks(blok, ['uninstalled'], [
                                'required', 'optional', 'conditional']))
                    elif state == 'toupdate':
                        self.apply_state(blok, state, ['installed'])
                        upgrade_state_bloks(state)(
                            self.get_bloks(blok, ['installed'], [
                                'required_by', 'optional_by',
                                'conditional_by']))
                    elif state == 'touninstall':
                        if Blok.check_if_the_conditional_are_installed(blok):
                            raise RegistryException(
                                "the blok %r can not be unistalled because "
                                "this blok is a conditional blok and all the "
                                "bloks in his conditional list are installed "
                                "You must uninstall one of them" % blok)
                        self.apply_state(blok, state, ['installed'])
                        upgrade_state_bloks(state)(self.get_bloks(blok, [
                            'installed', 'toinstall', 'touninstall'], [
                            'required_by', 'conditional_by']))
                        upgrade_state_bloks('toupdate')(self.get_bloks(blok, [
                            'installed', 'toinstall', 'touninstall'], [
                            'optional_by']))

            return wrap

        upgrade_state_bloks('touninstall')(uninstall or [])
        upgrade_state_bloks('toinstall')(install or [])
        upgrade_state_bloks('toupdate')(update or [])
        self.reload()
        self.expire_all()

    @log(logger, level='debug')
    def update_blok_list(self):
        if not self.blok_list_is_loaded:
            self.System.Blok.update_list()
            self.blok_list_is_loaded = True



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok.declarations import Declarations, hybrid_method
from anyblok.column import Integer

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin
Core = Declarations.Core


class TestHybridMethod(DBTestCase):

    def check_hybrid_method(self, Test):
        t1 = Test.insert(val=1)
        t2 = Test.insert(val=2)
        self.assertEqual(t1.val_is(1), True)
        self.assertEqual(t1.val_is(2), False)
        self.assertEqual(t2.val_is(1), False)
        self.assertEqual(t2.val_is(2), True)
        query = Test.query().filter(Test.val_is(1))
        self.assertEqual(query.count(), 1)
        self.assertEqual(query.first(), t1)
        query = Test.query().filter(Test.val_is(2))
        self.assertEqual(query.count(), 1)
        self.assertEqual(query.first(), t2)

    def test_hybrid_method_model(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                val = Integer(nullable=False)

                @hybrid_method
                def val_is(self, val):
                    return self.val == val

        registry = self.init_registry(add_in_registry)
        self.check_hybrid_method(registry.Test)

    def test_hybrid_method_model2(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                val = Integer(nullable=False)

                # check decorator with ()
                @hybrid_method()
                def val_is(self, val):
                    return self.val == val

        registry = self.init_registry(add_in_registry)
        self.check_hybrid_method(registry.Test)

    def test_hybrid_method_core(self):

        def add_in_registry():

            @register(Core)
            class SqlBase:

                @hybrid_method
                def val_is(self, val):
                    return self.val == val

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                val = Integer(nullable=False)

        registry = self.init_registry(add_in_registry)
        self.check_hybrid_method(registry.Test)

    def test_hybrid_method_mixin(self):

        def add_in_registry():

            @register(Mixin)
            class MTest:
                id = Integer(primary_key=True)
                val = Integer(nullable=False)

                @hybrid_method
                def val_is(self, val):
                    return self.val == val

            @register(Model)
            class Test(Mixin.MTest):
                pass

        registry = self.init_registry(add_in_registry)
        self.check_hybrid_method(registry.Test)

    def add_inherited_hybrid_method(self, withcore=False, withmixin=False,
                                    withmodel=False):

        @register(Core)
        class SqlBase:

            if withcore:
                @hybrid_method
                def val_is(self, val):
                    pass

        @register(Mixin)
        class MTest:
            id = Integer(primary_key=True)
            val = Integer(nullable=False)

            if withmixin:
                @hybrid_method
                def val_is(self, val):
                    pass

        @register(Model)
        class Test(Mixin.MTest):

            if withmodel:
                @hybrid_method
                def val_is(self, val):
                    pass

        @register(Model)  # noqa
        class Test:

            def val_is(self, val):
                return self.val == val

    def test_inherit_core(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withcore=True)
        self.check_hybrid_method(registry.Test)

    def test_inherit_mixin(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withmixin=True)
        self.check_hybrid_method(registry.Test)

    def test_inherit_model(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withmodel=True)
        self.check_hybrid_method(registry.Test)

    def test_inherit_core_and_mixin(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withcore=True, withmixin=True)
        self.check_hybrid_method(registry.Test)

    def test_inherit_core_and_model(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withcore=True, withmodel=True)
        self.check_hybrid_method(registry.Test)

    def test_inherit_mixin_and_model(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withmixin=True, withmodel=True)
        self.check_hybrid_method(registry.Test)

    def test_inherit_core_and_mixin_and_model(self):
        registry = self.init_registry(self.add_inherited_hybrid_method,
                                      withcore=True, withmixin=True,
                                      withmodel=True)
        self.check_hybrid_method(registry.Test)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from sqlalchemy import Sequence as SQLASequence
from anyblok.column import Integer, String


register = Declarations.register
System = Declarations.Model.System


@register(System)
class Sequence:
    """ System sequence """

    _cls_seq_name = 'system_sequence_seq_name'

    id = Integer(primary_key=True)
    code = String(nullable=False)
    number = Integer(nullable=False)
    seq_name = String(nullable=False)
    formater = String(nullable=False, default="{seq}")

    @classmethod
    def initialize_model(cls):
        """ Create the sequence to determine name """
        super(Sequence, cls).initialize_model()
        seq = SQLASequence(cls._cls_seq_name)
        seq.create(cls.registry.bind)

        if hasattr(cls.registry, '_need_sequence_to_create_if_not_exist'):
            if cls.registry._need_sequence_to_create_if_not_exist:
                for vals in cls.registry._need_sequence_to_create_if_not_exist:
                    if 'formater' in vals and vals['formater'] is None:
                        del vals['formater']

                    if cls.query().filter(cls.code == vals['code']).count():
                        continue

                    cls.insert(**vals)

    @classmethod
    def create_sequence(cls, values):
        """ create the sequence for one instance """
        if 'seq_name' in values:
            seq_name = values['seq_name']
        else:
            seq_id = cls.registry.execute(SQLASequence(cls._cls_seq_name))
            seq_name = '%s_%d' % (cls.__tablename__, seq_id)
            values['seq_name'] = seq_name
        if 'number' in values:
            seq = SQLASequence(seq_name, values['number'])
        else:
            values['number'] = 0
            seq = SQLASequence(seq_name)

        seq.create(cls.registry.bind)
        return values

    @classmethod
    def insert(cls, **kwargs):
        """ Overwrite insert """
        return super(Sequence, cls).insert(**cls.create_sequence(kwargs))

    @classmethod
    def multi_insert(cls, *args):
        """ Overwrite multi_insert """
        res = [cls.create_sequence(x) for x in args]
        return super(Sequence, cls).multi_insert(*res)

    def nextval(self):
        """ return the next value of the sequence """
        nextval = self.registry.execute(SQLASequence(self.seq_name))
        self.update(number=nextval)
        return self.formater.format(code=self.code, seq=nextval, id=self.id)

    @classmethod
    def nextvalBy(cls, **kwargs):
        """ Get the first sequence filtering by entries and return the next
        value """
        filters = [getattr(cls, k) == v for k, v in kwargs.items()]
        query = cls.query().filter(*filters)
        if query.count():
            seq = query.first()
            return seq.nextval()

        return None



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.orm.session import object_state
from .exceptions import (
    ForbidUpdateException, ForbidDeleteException)
from anyblok import Declarations
from anyblok.common import anyblok_column_prefix

Mixin = Declarations.Mixin


@Declarations.register(Mixin)
class ForbidUpdate:

    def get_modified_fields(self):
        state = object_state(self)
        modified_fields = []
        for attr in state.manager.attributes:
            if not hasattr(attr.impl, 'get_history'):
                continue

        added, unmodified, deleted = attr.impl.get_history(
            state, state.dict)

        if added or deleted:
            field = attr.key
            if field.startswith(anyblok_column_prefix):
                field = field[len(anyblok_column_prefix):]

            modified_fields.append(field)

        return modified_fields

    @classmethod
    def before_update_orm_event(cls, mapper, connection, target):
        modified_fields = target.get_modified_fields()
        raise ForbidUpdateException(
            "The modification of %r on %s with field(s) %r is forbidden" % (
                target, cls.__registry_name__, modified_fields))


@Declarations.register(Mixin)
class ForbidDelete:

    @classmethod
    def before_delete_orm_event(cls, mapper, connection, target):
        raise ForbidDeleteException(
            "The deletion of %r on %r is forbidden" % (
                target, cls.__registry_name__))


@Declarations.register(Declarations.Mixin)
class ReadOnly(Mixin.ForbidUpdate, Mixin.ForbidDelete):
    pass



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.ext.hybrid import hybrid_method
from .plugins import ModelPluginBase


class HybridMethodPlugin(ModelPluginBase):

    def initialisation_tranformation_properties(self, properties,
                                                transformation_properties):
        """ Initialise the transform properties: hybrid_method

        :param properties: the properties declared in the model
        :param new_type_properties: param to add in a new base if need
        """
        if 'hybrid_method' not in transformation_properties:
            transformation_properties['hybrid_method'] = []

    def transform_base_attribute(self, attr, method, namespace, base,
                                 transformation_properties,
                                 new_type_properties):
        """ Find the sqlalchemy hybrid methods in the base to save the
        namespace and the method in the registry

        :param attr: attribute name
        :param method: method pointer of the attribute
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """
        if not hasattr(method, 'is_an_hybrid_method'):
            return
        elif method.is_an_hybrid_method is True:
            if attr not in transformation_properties['hybrid_method']:
                transformation_properties['hybrid_method'].append(attr)

    def insert_in_bases(self, new_base, namespace, properties,
                        transformation_properties):
        """ Create overload to define the write declaration of sqlalchemy
        hybrid method, add the overload in the declared bases of the
        namespace

        :param new_base: the base to be put on front of all bases
        :param namespace: the namespace of the model
        :param properties: the properties declared in the model
        :param transformation_properties: the properties of the model
        """
        type_properties = {}

        def apply_wrapper(attr):

            def wrapper(self, *args, **kwargs):
                self_ = self.registry.loaded_namespaces[self.__registry_name__]
                if self is self_:
                    return getattr(super(new_base, self), attr)(
                        self, *args, **kwargs)
                else:
                    return getattr(super(new_base, self), attr)(
                        *args, **kwargs)

            setattr(new_base, attr, hybrid_method(wrapper))

        if transformation_properties['hybrid_method']:
            for attr in transformation_properties['hybrid_method']:
                apply_wrapper(attr)

        return type_properties
